# Copyright 2015-2018 Lauri Himanen, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import absolute_import
from builtins import str
import re
import numpy as np
import logging
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.caching_backend import CachingLevel
from nomadcore.unit_conversion.unit_conversion import convert_unit
from nomadcore.baseclasses import CommonParser
from .inputparser import CP2KInputParser
from collections import defaultdict
logger = logging.getLogger("nomad")


class CP2KCommonParser(CommonParser):
    """
    This class is used to store and instantiate common parts of the
    hierarchical SimpleMatcher structure used in the parsing of a CP2K
    output file.
    """
    def __init__(self, parser_context):
        super(CP2KCommonParser, self).__init__(parser_context)
        self.section_method_index = None
        self.section_system_index = None
        self.test_electronic_structure_method = "DFT"
        self.basis_to_kind_mapping = []
        self.atom_kind_info = defaultdict(dict)  # Map from kind number to kind information
        self.basis_set_info = defaultdict(dict)  # Map from kind number to basis set information

        #=======================================================================
        # Cache levels
        self.caching_levels = {
            'x_cp2k_atoms': CachingLevel.ForwardAndCache,
            'section_XC_functionals': CachingLevel.ForwardAndCache,
            'self_interaction_correction_method': CachingLevel.Cache,
            'x_cp2k_section_program_information': CachingLevel.ForwardAndCache,
            'x_cp2k_section_quickstep_settings': CachingLevel.ForwardAndCache,
            'x_cp2k_section_atomic_kind': CachingLevel.ForwardAndCache,
            'x_cp2k_section_kind_basis_set': CachingLevel.ForwardAndCache,
        }

        #=======================================================================
        # Globally cached values
        self.cache_service.add("simulation_cell", single=False, update=False)
        self.cache_service.add("lattice_vectors", single=False, update=False)
        self.cache_service.add("number_of_scf_iterations", 0)
        self.cache_service.add("atom_positions", single=False, update=True)
        self.cache_service.add("atom_labels", single=False, update=False)
        self.cache_service.add("number_of_atoms", single=False, update=False)
        self.cache_service.add("basis_set_kind", single=False, update=False)
        self.cache_service.add("basis_set_name", single=False, update=False)
        self.cache_service.add("basis_set_planewave_cutoff", update=False)
        self.cache_service.add("mapping_section_basis_set_cell_dependent", single=False, update=False)
        self.cache_service.add("map_kind_to_basis", single=False, update=False)
        self.cache_service.add("map_index_to_kind", single=False, update=False)
        self.cache_service.add("map_kind_number_to_basis_ref", single=False, update=False)
        self.cache_service.add("electron_density_filename", single=False, update=True)

    #===========================================================================
    # SimpleMatchers

    # SimpleMatcher for the header that is common to all run types
    def header(self):
        return SM( " DBCSR\| Multiplication driver",
            forwardMatch=True,
            subMatchers=[
                SM( " DBCSR\| Multiplication driver",
                    forwardMatch=True,
                    sections=['x_cp2k_section_dbcsr'],
                    subMatchers=[
                        SM( " DBCSR\| Multiplication driver\s+(?P<x_cp2k_dbcsr_multiplication_driver>{})".format(self.regexs.word)),
                        SM( " DBCSR\| Multrec recursion limit\s+(?P<x_cp2k_dbcsr_multrec_recursion_limit>{})".format(self.regexs.int)),
                        SM( " DBCSR\| Multiplication stack size\s+(?P<x_cp2k_dbcsr_multiplication_stack_size>{})".format(self.regexs.int)),
                        SM( " DBCSR\| Multiplication size stacks\s+(?P<x_cp2k_dbcsr_multiplication_size_stacks>{})".format(self.regexs.int)),
                        SM( " DBCSR\| Use subcommunicators\s+(?P<x_cp2k_dbcsr_use_subcommunicators>{})".format(self.regexs.letter)),
                        SM( " DBCSR\| Use MPI combined types\s+(?P<x_cp2k_dbcsr_use_mpi_combined_types>{})".format(self.regexs.letter)),
                        SM( " DBCSR\| Use MPI memory allocation\s+(?P<x_cp2k_dbcsr_use_mpi_memory_allocation>{})".format(self.regexs.letter)),
                        SM( " DBCSR\| Use Communication thread\s+(?P<x_cp2k_dbcsr_use_communication_thread>{})".format(self.regexs.letter)),
                        SM( " DBCSR\| Communication thread load\s+(?P<x_cp2k_dbcsr_communication_thread_load>{})".format(self.regexs.int)),
                    ]
                ),
                SM( "  **** **** ******  **  PROGRAM STARTED AT".replace("*", "\*"),
                    forwardMatch=True,
                    sections=['x_cp2k_section_startinformation'],
                    subMatchers=[
                        SM( "  **** **** ******  **  PROGRAM STARTED AT\s+(?P<x_cp2k_start_time>{})".replace("*", "\*").format(self.regexs.eol)),
                        SM( " ***** ** ***  *** **   PROGRAM STARTED ON\s+(?P<x_cp2k_start_host>{})".replace("*", "\*").format(self.regexs.word)),
                        SM( " **    ****   ******    PROGRAM STARTED BY\s+(?P<x_cp2k_start_user>{})".replace("*", "\*").format(self.regexs.word)),
                        SM( " ***** **    ** ** **   PROGRAM PROCESS ID\s+(?P<x_cp2k_start_id>{})".replace("*", "\*").format(self.regexs.int)),
                        SM( "  **** **  *******  **  PROGRAM STARTED IN".replace("*", "\*"),
                            forwardMatch=True,
                            adHoc=self.adHoc_run_dir("x_cp2k_start_path"),
                        )
                    ]
                ),
                SM( " CP2K\| version string:",
                    sections=['x_cp2k_section_program_information'],
                    forwardMatch=True,
                    subMatchers=[
                        SM( " CP2K\| version string:\s+(?P<program_version>{})".format(self.regexs.eol)),
                        SM( " CP2K\| source code revision number:\s+svn:(?P<x_cp2k_svn_revision>\d+)"),
                        SM( " CP2K\| is freely available from{}".format(self.regexs.eol)),
                        SM( " CP2K\| Program compiled at\s+(?P<x_cp2k_program_compilation_datetime>{})".format(self.regexs.eol)),
                        SM( " CP2K\| Program compiled on\s+(?P<program_compilation_host>{})".format(self.regexs.eol)),
                        SM( " CP2K\| Program compiled for{}".format(self.regexs.eol)),
                        SM( " CP2K\| Input file name\s+(?P<x_cp2k_input_filename>{})".format(self.regexs.eol)),
                    ]
                ),
                SM( " GLOBAL\|",
                    sections=['x_cp2k_section_global_settings'],
                    subMatchers=[
                        SM( " GLOBAL\| Force Environment number"),
                        SM( " GLOBAL\| Basis set file name\s+(?P<x_cp2k_basis_set_filename>{})".format(self.regexs.eol)),
                        SM( " GLOBAL\| Geminal file name\s+(?P<x_cp2k_geminal_filename>{})".format(self.regexs.eol)),
                        SM( " GLOBAL\| Potential file name\s+(?P<x_cp2k_potential_filename>{})".format(self.regexs.eol)),
                        SM( " GLOBAL\| MM Potential file name\s+(?P<x_cp2k_mm_potential_filename>{})".format(self.regexs.eol)),
                        SM( " GLOBAL\| Coordinate file name\s+(?P<x_cp2k_coordinate_filename>{})".format(self.regexs.eol)),
                        SM( " GLOBAL\| Method name\s+(?P<x_cp2k_method_name>{})".format(self.regexs.eol)),
                        SM( " GLOBAL\| Project name"),
                        SM( " GLOBAL\| Preferred FFT library\s+(?P<x_cp2k_preferred_fft_library>{})".format(self.regexs.eol)),
                        SM( " GLOBAL\| Preferred diagonalization lib.\s+(?P<x_cp2k_preferred_diagonalization_library>{})".format(self.regexs.eol)),
                        SM( " GLOBAL\| Run type\s+(?P<x_cp2k_run_type>{})".format(self.regexs.eol)),
                        SM( " GLOBAL\| All-to-all communication in single precision"),
                        SM( " GLOBAL\| FFTs using library dependent lengths"),
                        SM( " GLOBAL\| Global print level"),
                        SM( " GLOBAL\| Total number of message passing processes"),
                        SM( " GLOBAL\| Number of threads for this process"),
                        SM( " GLOBAL\| This output is from process"),
                    ],
                ),
                SM( " CELL\|",
                    adHoc=self.adHoc_x_cp2k_section_cell,
                ),
            ]
        )

    # SimpleMatcher for the restart information
    def restart(self):
        return SM( re.escape(" *                            RESTART INFORMATION                              *"),
            sections=["x_cp2k_section_restart_information"],
            subMatchers=[
                SM( re.escape(" *******************************************************************************")),
                SM( re.escape(" *                                                                             *")),
                SM( re.escape(" *    RESTART FILE NAME: (?P<x_cp2k_restart_file_name>{})\s+*".format(self.regexs.word))),
                SM( re.escape(" *                                                                             *")),
                SM( re.escape(" * RESTARTED QUANTITIES:                                                       *")),
                SM( re.escape(" *                       (?P<x_cp2k_restarted_quantity>{})\s+*".format(self.regexs.word)),
                    repeats=True,
                ),
                SM( re.escape(" *******************************************************************************"))
            ]
        )

    # SimpleMatcher for the footer that is common to all run types
    def footer(self):
        return SM( " -                                DBCSR STATISTICS                             -",
            forwardMatch=True,
            subMatchers=[
                SM( re.escape("  **** **** ******  **  PROGRAM ENDED AT"),
                    forwardMatch=True,
                    sections=['x_cp2k_section_end_information'],
                    subMatchers=[
                        SM( "  **** **** ******  **  PROGRAM ENDED AT\s+(?P<x_cp2k_end_time>{})".replace("*", "\*").format(self.regexs.eol)),
                        SM( " ***** ** ***  *** **   PROGRAM RAN ON\s+(?P<x_cp2k_end_host>{})".replace("*", "\*").format(self.regexs.word)),
                        SM( " **    ****   ******    PROGRAM RAN BY\s+(?P<x_cp2k_end_user>{})".replace("*", "\*").format(self.regexs.word)),
                        SM( " ***** **    ** ** **   PROGRAM PROCESS ID\s+(?P<x_cp2k_end_id>{})".replace("*", "\*").format(self.regexs.int)),
                        SM( "  **** **  *******  **  PROGRAM STOPPED IN".replace("*", "\*"),
                            forwardMatch=True,
                            adHoc=self.adHoc_run_dir("x_cp2k_end_path"),
                        )
                    ]
                ),
            ]
        )

    # SimpleMatcher for an SCF wavefunction optimization
    def quickstep_calculation(self):
        return SM( " SCF WAVEFUNCTION OPTIMIZATION",
            sections=["x_cp2k_section_quickstep_calculation"],
            subMatchers=[
                SM( r"  Trace\(PS\):",
                    sections=["x_cp2k_section_scf_iteration"],
                    repeats=True,
                    subMatchers=[
                        SM( r"  Exchange-correlation energy:\s+(?P<x_cp2k_energy_XC_scf_iteration__hartree>{})".format(self.regexs.float)),
                        SM( r"\s+\d+\s+\S+\s+{0}\s+{0}\s+{0}\s+(?P<x_cp2k_energy_total_scf_iteration__hartree>{0})\s+(?P<x_cp2k_energy_change_scf_iteration__hartree>{0})".format(self.regexs.float)),
                    ]
                ),
                SM( r"  \*\*\* SCF run converged in\s+(\d+) steps \*\*\*",
                    adHoc=self.adHoc_single_point_converged
                ),
                SM( r"  \*\*\* SCF run NOT converged \*\*\*",
                    adHoc=self.adHoc_single_point_not_converged
                ),
                SM( r" The electron density is written in cube file format to the file:",
                    subMatchers=[
                        SM(""),
                        SM(
                            "\s+(.+\.cube)",
                            startReAction=self.startReAction_save_cube_filename,
                        ),
                    ]
                ),
                SM( r"  Electronic kinetic energy:\s+(?P<x_cp2k_electronic_kinetic_energy__hartree>{})".format(self.regexs.float)),
                SM( r" **************************** NUMERICAL STRESS ********************************".replace("*", "\*"),
                    # endReStr=" **************************** NUMERICAL STRESS END *****************************".replace("*", "\*"),
                    adHoc=self.adHoc_stress_calculation,
                ),
                SM( r" ENERGY\| Total FORCE_EVAL \( \w+ \) energy \(a\.u\.\):\s+(?P<x_cp2k_energy_total__hartree>{0})".format(self.regexs.float),
                ),
                SM( r" ATOMIC FORCES in \[a\.u\.\]"),
                SM( r" # Atom   Kind   Element          X              Y              Z",
                    adHoc=self.adHoc_atom_forces,
                ),
                SM( r" (?:NUMERICAL )?STRESS TENSOR \[GPa\]",
                    sections=["x_cp2k_section_stress_tensor"],
                    subMatchers=[
                        SM( r"\s+X\s+Y\s+Z",
                            adHoc=self.adHoc_stress_tensor,
                        ),
                        SM( "  1/3 Trace\(stress tensor\):\s+(?P<x_cp2k_stress_tensor_one_third_of_trace__GPa>{})".format(self.regexs.float)),
                        SM( "  Det\(stress tensor\)\s+:\s+(?P<x_cp2k_stress_tensor_determinant__GPa3>{})".format(self.regexs.float)),
                        SM( " EIGENVECTORS AND EIGENVALUES OF THE STRESS TENSOR",
                            adHoc=self.adHoc_stress_tensor_eigenpairs),
                    ]
                )
            ]
        )

    # SimpleMatcher the stuff that is done to initialize a quickstep calculation
    def quickstep_header(self):
        return SM( re.escape(" **                                                ... make the atoms dance   **"),
            forwardMatch=True,
            sections=["x_cp2k_section_quickstep_settings"],
            subMatchers=[
                SM( " DFT\|",
                    forwardMatch=True,
                    subMatchers=[
                        SM( " DFT\| Spin restricted Kohn-Sham (RKS) calculation\s+(?P<x_cp2k_spin_restriction>{})".format(self.regexs.word)),
                        SM( " DFT\| Multiplicity\s+(?P<spin_target_multiplicity>{})".format(self.regexs.int)),
                        SM( " DFT\| Number of spin states\s+(?P<number_of_spin_channels>{})".format(self.regexs.int)),
                        SM( " DFT\| Charge\s+(?P<total_charge>{})".format(self.regexs.int)),
                        SM( " DFT\| Self-interaction correction \(SIC\)\s+(?P<self_interaction_correction_method>[^\n]+)"),
                    ],
                ),
                SM( " vdW POTENTIAL\|\s+",
                    forwardMatch=True,
                    sections=["x_cp2k_section_vdw_settings"],
                    subMatchers=[
                        SM( " vdW POTENTIAL\|\s+(?P<x_cp2k_vdw_type>{})".format(self.regexs.eol)),
                        SM( " vdW POTENTIAL\|\s+Potential Form:\s+(?P<x_cp2k_vdw_name>{})".format(self.regexs.eol)),
                        SM( " vdW POTENTIAL\|\s+BJ Damping:\s+(?P<x_cp2k_vdw_bj_damping_name>{})".format(self.regexs.eol)),
                        SM( " vdW POTENTIAL\|\s+Cutoff Radius \[Bohr\]:\s+(?P<x_cp2k_vdw_cutoff_radius>{})".format(self.regexs.float)),
                        SM( " vdW POTENTIAL\|\+sScaling Factor:s+(?P<x_cp2k_vdw_scaling_factor>{})".format(self.regexs.float),
                            sections=["x_cp2k_section_vdw_d2_settings"],
                            subMatchers=[
                                SM( " vdW POTENTIAL\|\s+Exp Prefactor for Damping:\s+(?P<x_cp2k_vdw_damping_factor>{})".format(self.regexs.float)),
                                SM( " vdW PARAMETER\|\s+Atom=(?P<x_cp2k_vdw_parameter_element_name>{})\s+C6\[J*nm^6*mol^-1\]=\s+(?P<x_cp2k_vdw_parameter_c6>{})\s+r\(vdW\)\[A\]=\s+(?P<x_cp2k_vdw_parameter_radius>{})".format(self.regexs.word, self.regexs.float, self.regexs.float),
                                    sections=["x_cp2k_section_vdw_element_settings"],
                                    repeats=True,
                                ),
                            ],
                        ),
                        SM( " vdW POTENTIAL\|\s+s6 Scaling Factor:\s+(?P<x_cp2k_vdw_s6_scaling_factor>{})".format(self.regexs.float),
                            sections=["x_cp2k_section_vdw_d3_settings"],
                            subMatchers=[
                                SM( " vdW POTENTIAL\|\s+sr6 Scaling Factor:\s+(?P<x_cp2k_vdw_sr6_scaling_factor>{})".format(self.regexs.float)),
                                SM( " vdW POTENTIAL\|\s+s8 Scaling Factor:\s+(?P<x_cp2k_vdw_s8_scaling_factor>{})".format(self.regexs.float)),
                                SM( " vdW POTENTIAL\|\s+Cutoff for CN calculation:\s+(?P<x_cp2k_vdw_cn_cutoff>{})".format(self.regexs.float)),
                            ],
                        )
                    ]
                ),
                SM( " DFT\+U\|",
                    adHoc=self.adHoc_dft_plus_u,
                ),
                SM( " QS\|",
                    forwardMatch=True,
                    subMatchers=[
                        SM( " QS\| Method:\s+(?P<x_cp2k_quickstep_method>{})".format(self.regexs.word)),
                        SM( " QS\| Density plane wave grid type\s+{}".format(self.regexs.eol)),
                        SM( " QS\| Number of grid levels:\s+{}".format(self.regexs.int)),
                        SM( " QS\| Density cutoff \[a\.u\.\]:\s+(?P<x_cp2k_planewave_cutoff>{})".format(self.regexs.float)),
                        SM( " QS\| Multi grid cutoff \[a\.u\.\]: 1\) grid level\s+{}".format(self.regexs.float)),
                        SM( " QS\|                           2\) grid level\s+{}".format(self.regexs.float)),
                        SM( " QS\|                           3\) grid level\s+{}".format(self.regexs.float)),
                        SM( " QS\|                           4\) grid level\s+{}".format(self.regexs.float)),
                        SM( " QS\| Grid level progression factor:\s+{}".format(self.regexs.float)),
                        SM( " QS\| Relative density cutoff \[a\.u\.\]:".format(self.regexs.float)),
                        SM( " QS\| Consistent realspace mapping and integration"),
                        SM( " QS\| Interaction thresholds: eps_pgf_orb:\s+{}".format(self.regexs.float)),
                        SM( " QS\|                         eps_filter_matrix:\s+{}".format(self.regexs.float)),
                        SM( " QS\|                         eps_core_charge:\s+{}".format(self.regexs.float)),
                        SM( " QS\|                         eps_rho_gspace:\s+{}".format(self.regexs.float)),
                        SM( " QS\|                         eps_rho_rspace:\s+{}".format(self.regexs.float)),
                        SM( " QS\|                         eps_gvg_rspace:\s+{}".format(self.regexs.float)),
                        SM( " QS\|                         eps_ppl:\s+{}".format(self.regexs.float)),
                        SM( " QS\|                         eps_ppnl:\s+{}".format(self.regexs.float)),
                    ],
                ),
                SM( " ATOMIC KIND INFORMATION",
                    sections=["x_cp2k_section_atomic_kinds"],
                    subMatchers=[
                        SM( "\s+(?P<x_cp2k_kind_number>{0})\. Atomic kind: (?P<x_cp2k_kind_label>{1})\s+Number of atoms:\s+(?P<x_cp2k_kind_number_of_atoms>{1})".format(self.regexs.int, self.regexs.word),
                            repeats=True,
                            sections=["x_cp2k_section_atomic_kind", "x_cp2k_section_kind_basis_set"],
                            subMatchers=[
                                SM( "     Orbital Basis Set\s+(?P<x_cp2k_kind_basis_set_name>{})".format(self.regexs.word)),
                                SM( "       Number of orbital shell sets:\s+(?P<x_cp2k_basis_set_number_of_orbital_shell_sets>{})".format(self.regexs.int)),
                                SM( "       Number of orbital shells:\s+(?P<x_cp2k_basis_set_number_of_orbital_shells>{})".format(self.regexs.int)),
                                SM( "       Number of primitive Cartesian functions:\s+(?P<x_cp2k_basis_set_number_of_primitive_cartesian_functions>{})".format(self.regexs.int)),
                                SM( "       Number of Cartesian basis functions:\s+(?P<x_cp2k_basis_set_number_of_cartesian_basis_functions>{})".format(self.regexs.int)),
                                SM( "       Number of spherical basis functions:\s+(?P<x_cp2k_basis_set_number_of_spherical_basis_functions>{})".format(self.regexs.int)),
                                SM( "       Norm type:\s+(?P<x_cp2k_basis_set_norm_type>{})".format(self.regexs.int)),
                            ]
                        )
                    ]
                ),
                SM( "  Total number of",
                    forwardMatch=True,
                    sections=["x_cp2k_section_total_numbers"],
                    subMatchers=[
                        SM( "  Total number of            - Atomic kinds:\s+(?P<x_cp2k_atomic_kinds>\d+)"),
                        SM( "\s+- Atoms:\s+(?P<x_cp2k_atoms>\d+)"),
                        SM( "\s+- Shell sets:\s+(?P<x_cp2k_shell_sets>\d+)"),
                        SM( "\s+- Shells:\s+(?P<x_cp2k_shells>\d+)"),
                        SM( "\s+- Primitive Cartesian functions:\s+(?P<x_cp2k_primitive_cartesian_functions>\d+)"),
                        SM( "\s+- Cartesian basis functions:\s+(?P<x_cp2k_cartesian_basis_functions>\d+)"),
                        SM( "\s+- Spherical basis functions:\s+(?P<x_cp2k_spherical_basis_functions>\d+)"),
                    ]
                ),
                SM( " Maximum angular momentum of",
                    forwardMatch=True,
                    sections=["x_cp2k_section_maximum_angular_momentum"],
                    subMatchers=[
                        SM( "  Maximum angular momentum of- Orbital basis functions::\s+(?P<x_cp2k_orbital_basis_functions>\d+)"),
                        SM( "\s+- Local part of the GTH pseudopotential:\s+(?P<x_cp2k_local_part_of_gth_pseudopotential>\d+)"),
                        SM( "\s+- Non-local part of the GTH pseudopotential:\s+(?P<x_cp2k_non_local_part_of_gth_pseudopotential>\d+)"),
                    ]
                ),
                SM( " MODULE QUICKSTEP:  ATOMIC COORDINATES IN angstrom",
                    forwardMatch=True,
                    subMatchers=[
                        SM( " MODULE QUICKSTEP:  ATOMIC COORDINATES IN angstrom",
                            adHoc=self.adHoc_x_cp2k_section_quickstep_atom_information(),
                        )
                    ]
                ),
                SM( " SCF PARAMETERS",
                    forwardMatch=True,
                    subMatchers=[
                        SM( " SCF PARAMETERS         Density guess:\s+{}".format(self.regexs.eol)),
                        SM( "                        max_scf:\s+(?P<scf_max_iteration>{})".format(self.regexs.int)),
                        SM( "                        max_scf_history:\s+{}".format(self.regexs.int)),
                        SM( "                        max_diis:\s+{}".format(self.regexs.int)),
                        SM( "                        eps_scf:\s+(?P<scf_threshold_energy_change__hartree>{})".format(self.regexs.float)),
                    ]
                ),
                SM( " MP2\|",
                    adHoc=self.adHoc_mp2
                ),
                SM( " RI-RPA\|",
                    adHoc=self.adHoc_rpa
                ),
            ]
        )

    #===========================================================================
    # onClose triggers
    def onClose_x_cp2k_section_total_numbers(self, backend, gIndex, section):
        """Keep track of how many SCF iteration are made."""
        number_of_atoms = section.get_latest_value("x_cp2k_atoms")
        if number_of_atoms is not None:
            self.cache_service["number_of_atoms"] = number_of_atoms

    # def onClose_x_cp2k_section_quickstep_calculation(self, backend, gIndex, section):
        # print "quickstep CLOSED"

    # def onClose_x_cp2k_section_geometry_optimization_step(self, backend, gIndex, section):
        # print "Optimisation step CLOSED"

    def onClose_section_method(self, backend, gIndex, section):
        """When all the functional definitions have been gathered, matches them
        with the nomad correspondents and combines into one single string which
        is put into the backend.
        """
        self.section_method_index = gIndex

        # Transform the CP2K self-interaction correction string to the NOMAD
        # correspondent, and push directly to the superBackend to avoid caching
        try:
            sic_cp2k = section.get_latest_value("self_interaction_correction_method")
            if sic_cp2k is not None:
                sic_map = {
                    "NO": "",
                    "AD SIC": "SIC_AD",
                    "Explicit Orbital SIC": "SIC_EXPLICIT_ORBITALS",
                    "SPZ/MAURI SIC": "SIC_MAURI_SPZ",
                    "US/MAURI SIC": "SIC_MAURI_US",
                }
                sic_nomad = sic_map.get(sic_cp2k)
                if sic_nomad is not None:
                    backend.superBackend.addValue('self_interaction_correction_method', sic_nomad)
                else:
                    logger.warning("Unknown self-interaction correction method used.")
        except:
            pass

    def onClose_section_run(self, backend, gIndex, section):
        backend.addValue("program_name", "CP2K")

    def onClose_x_cp2k_section_quickstep_settings(self, backend, gIndex, section):
        backend.addValue("program_basis_set_type", "gaussians")
        backend.addValue("electronic_structure_method", self.test_electronic_structure_method)

        # Collect the atomic kind information and push it to backend
        kind_ids = {}
        for kind_number, info in self.atom_kind_info.items():
            kindID = backend.openSection("section_method_atom_kind")
            kind_ids[kind_number] = kindID
            label = info["label"]
            atom_number = info.get("element_number")
            if atom_number is not None:
                backend.addValue("method_atom_kind_atom_number", atom_number)
            backend.addValue("method_atom_kind_label", label)
            backend.closeSection("section_method_atom_kind", kindID)

        # Cell dependent basis information
        ryd_cutoff = section.get_latest_value("x_cp2k_planewave_cutoff")
        if ryd_cutoff is not None:
            gid = backend.openSection("section_basis_set_cell_dependent")
            self.cache_service["mapping_section_basis_set_cell_dependent"] = gid
            ha_cutoff = convert_unit(2*ryd_cutoff, "hartree")
            backend.addValue("basis_set_planewave_cutoff", ha_cutoff)
            self.cache_service["basis_set_planewave_cutoff"] = ryd_cutoff
            backend.closeSection("section_basis_set_cell_dependent", gid)

        # Atom centered basis set information
        basis_ids = {}
        map_kind_number_to_basis_ref = {}
        for kind_number, info in self.basis_set_info.items():
            basis_section_id = backend.openSection("section_basis_set_atom_centered")
            basis_ids[kind_number] = basis_section_id
            map_kind_number_to_basis_ref[kind_number] = basis_section_id
            name = info["name"]
            atom_number = info.get("element_number")
            if atom_number is not None:
                backend.addValue("basis_set_atom_number", atom_number)
            backend.addValue("basis_set_atom_centered_short_name", name)
            backend.closeSection("section_basis_set_atom_centered", basis_section_id)
        self.cache_service["map_kind_number_to_basis_ref"] = map_kind_number_to_basis_ref

        # Add the basis infomation to section_method
        mapping = []
        dict_map = {}
        for kind_number, basis_id in basis_ids.items():
            kind_id = kind_ids[kind_number]
            mapping.append((basis_id, kind_id))
            dict_map[kind_id] = basis_id
        method_basis_id = backend.openSection("section_method_basis_set")
        if mapping:
            mapping = np.array(mapping)
            self.cache_service["map_kind_to_basis"] = dict_map
            backend.addArrayValues("mapping_section_method_basis_set_atom_centered", np.array(mapping))
        backend.addValue("method_basis_set_kind", "wavefunction")
        self.cache_service.addValue("mapping_section_method_basis_set_cell_associated")
        backend.addValue("number_of_basis_sets_atom_centered", len(self.basis_set_info))
        backend.closeSection("section_method_basis_set", method_basis_id)

    def onClose_x_cp2k_section_vdw_settings(self, backend, gIndex, section):
        """Figures out the common metainfo for vdw from the CP2K specific
        settings.
        """
        vdw_name = section["x_cp2k_vdw_name"][0]
        name_map = {
            "S. Grimme, JCC 27: 1787 (2006)": "G06",
            "S. Grimme et al, JCP 132: 154104 (2010)": "G10",
        }
        nomad_vdw_name = name_map.get(vdw_name)
        if nomad_vdw_name is not None:
            backend.addValue("van_der_Waals_method", nomad_vdw_name)

    def onClose_x_cp2k_section_atomic_kind(self, backend, gIndex, section):
        # basisID = backend.openSection("section_basis_set_atom_centered")

        # Save the kind labels. These wil be connected to atomic numbers later
        # on when the atomic numbers are listed in the atomic positions.
        kind_number = int(section.get_latest_value("x_cp2k_kind_number"))
        kind_label = section.get_latest_value("x_cp2k_kind_label")
        self.atom_kind_info[kind_number]["label"] = kind_label

        # Save all the basis set information for later use. They will be pushed
        # later when an atom number can be associated with the basis.
        basis_set_name = section.get_latest_value(["x_cp2k_section_kind_basis_set", "x_cp2k_kind_basis_set_name"])
        basis_info = self.basis_set_info[kind_number]
        basis_info["name"] = basis_set_name

    def onClose_x_cp2k_section_atomic_kinds(self, backend, gIndex, section):
        # Store the name and kind of the basis set for later use (stored inside
        # single_configuration_calculation).
        atomic_kinds = section["x_cp2k_section_atomic_kind"]
        long_basis_name = []
        for kind in atomic_kinds:
            kind_basis = kind["x_cp2k_section_kind_basis_set"][0]
            basis_name = kind_basis["x_cp2k_kind_basis_set_name"][0]
            long_basis_name.append(basis_name)
        self.cache_service["basis_set_kind"] = "wavefunction"
        self.cache_service["basis_set_name"] = "_".join(long_basis_name)

    def onClose_x_cp2k_section_program_information(self, backend, gIndex, section):
        input_file = section.get_latest_value("x_cp2k_input_filename")
        self.file_service.set_file_id(input_file, "input")

    def onClose_x_cp2k_section_global_settings(self, backend, gIndex, section):
        # If the input file is available, parse it
        filepath = self.file_service.get_file_by_id("input")
        if filepath is not None:
            input_parser = CP2KInputParser(self.parser_context)
            input_parser.parse(filepath)
        else:
            logger.warning("The input file of the calculation could not be found.")

    def onClose_section_system(self, backend, gIndex, section):
        """Stores the index of the section method. Should always be 0, but
        let's get it dynamically just in case there's something wrong.
        """
        self.section_system_index = gIndex
        self.cache_service.addValue("number_of_atoms")
        self.cache_service.addArrayValues("configuration_periodic_dimensions")
        self.cache_service.addArrayValues("atom_labels")

    def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
        # Write the references to section_method and section_system
        backend.addValue('single_configuration_to_calculation_method_ref', self.section_method_index)
        backend.addValue('single_configuration_calculation_to_system_ref', self.section_system_index)

        scc_basis_id = backend.openSection("section_basis_set")

        # Basis kind
        self.cache_service.addValue("basis_set_kind")

        # Basis name
        basis_name = self.cache_service["basis_set_name"]
        if basis_name is not None:
            cutoff = self.cache_service["basis_set_planewave_cutoff"]
            if cutoff is not None:
                basis_name += "_PW_{}".format(cutoff)
                backend.addValue("basis_set_name", basis_name)

        # Gaussian mapping
        map_index_to_basis = []
        map_kind_number_to_basis_ref = self.cache_service["map_kind_number_to_basis_ref"]
        map_index_to_kind = self.cache_service["map_index_to_kind"]
        # print(map_kind_number_to_basis_ref)
        # print(map_index_to_kind)

        if map_index_to_kind is not None and map_kind_number_to_basis_ref is not None:
            indices = map_index_to_kind.keys()
            for index in sorted(indices):
                kind = map_index_to_kind[index]
                basis_ref = map_kind_number_to_basis_ref[kind]
                map_index_to_basis.append(basis_ref)
            map_index_to_basis = np.array(map_index_to_basis)
            backend.addArrayValues("mapping_section_basis_set_atom_centered", map_index_to_basis)

        # Cell dependent basis mapping
        self.cache_service.addValue("mapping_section_basis_set_cell_dependent")

        # If a cube file was found, get the filepath. Currently no information
        # is read from this file because there is no suitable metainfo for it.
        filename = self.cache_service["electron_density_filename"]
        if filename is not None:
            self.parser_context.file_service.get_absolute_path_to_file(filename)

        backend.closeSection("section_basis_set", scc_basis_id)

    #===========================================================================
    # adHoc functions
    def adHoc_x_cp2k_section_cell(self, parser):
        """Used to extract the cell information.
        """
        # Read the lines containing the cell vectors
        a_line = parser.fIn.readline()
        b_line = parser.fIn.readline()
        c_line = parser.fIn.readline()

        # Define the regex that extracts the components and apply it to the lines
        regex_string = r" CELL\| Vector \w \[angstrom\]:\s+({0})\s+({0})\s+({0})".format(self.regexs.float)
        regex_compiled = re.compile(regex_string)
        a_result = regex_compiled.match(a_line)
        b_result = regex_compiled.match(b_line)
        c_result = regex_compiled.match(c_line)

        # Convert the string results into a 3x3 numpy array
        cell = np.zeros((3, 3))
        cell[0, :] = [float(x) for x in a_result.groups()]
        cell[1, :] = [float(x) for x in b_result.groups()]
        cell[2, :] = [float(x) for x in c_result.groups()]

        # Push the results to cache
        self.cache_service["simulation_cell"] = cell
        self.cache_service["lattice_vectors"] = cell

    def adHoc_atom_forces(self, parser):
        """Used to extract the final atomic forces printed at the end of a
        calculation.
        """
        end_str = " SUM OF ATOMIC FORCES"
        end = False
        force_array = []

        # Loop through coordinates until the sum of forces is read
        while not end:
            line = parser.fIn.readline()
            if line.startswith(end_str):
                end = True
            else:
                forces = line.split()[-3:]
                forces = [float(x) for x in forces]
                force_array.append(forces)
        force_array = np.array(force_array)

        # If anything found, push the results to the correct section
        if len(force_array) != 0:
            # self.cache_service["atom_forces"] = force_array
            self.backend.addArrayValues("x_cp2k_atom_forces", force_array, unit="hartree / bohr")

    def adHoc_stress_tensor(self, parser):
        """Used to extract the stress tensor printed at the end of a
        calculation.
        """
        row1 = [float(x) for x in parser.fIn.readline().split()[-3:]]
        row2 = [float(x) for x in parser.fIn.readline().split()[-3:]]
        row3 = [float(x) for x in parser.fIn.readline().split()[-3:]]
        stress_array = np.array([row1, row2, row3])
        parser.backend.addArrayValues("x_cp2k_stress_tensor", stress_array, unit="GPa")

    def adHoc_stress_calculation(self, parser):
        """Used to skip over the stress tensor calculation details.
        """
        end_line = " **************************** NUMERICAL STRESS END *****************************\n"
        finished = False
        while not finished:
            line = parser.fIn.readline()
            if line == end_line:
                finished = True

    def adHoc_stress_tensor_eigenpairs(self, parser):
        """Parses the stress tensor eigenpairs.
        """
        parser.fIn.readline()
        eigenvalues = np.array([float(x) for x in parser.fIn.readline().split()])
        parser.fIn.readline()
        row1 = [float(x) for x in parser.fIn.readline().split()]
        row2 = [float(x) for x in parser.fIn.readline().split()]
        row3 = [float(x) for x in parser.fIn.readline().split()]
        eigenvectors = np.array([row1, row2, row3])
        parser.backend.addArrayValues("x_cp2k_stress_tensor_eigenvalues", eigenvalues, unit="GPa")
        parser.backend.addArrayValues("x_cp2k_stress_tensor_eigenvectors", eigenvectors)

    def adHoc_single_point_converged(self, parser):
        """Called when the SCF cycle of a single point calculation has converged.
        """
        parser.backend.addValue("x_cp2k_quickstep_converged", True)

    def adHoc_single_point_not_converged(self, parser):
        """Called when the SCF cycle of a single point calculation did not converge.
        """
        parser.backend.addValue("x_cp2k_quickstep_converged", False)

    def adHoc_x_cp2k_section_quickstep_atom_information(self):
        """Used to extract the initial atomic coordinates and names in the
        Quickstep module.
        """
        def wrapper(parser):

            # Define the regex that extracts the information
            regex_string = r"\s+(\d+)\s+(\d+)\s+(\w+)\s+(\d+)\s+({0})\s+({0})\s+({0})".format(self.regexs.float)
            regex_compiled = re.compile(regex_string)

            match = True
            coordinates = []
            labels = []

            # Currently these three lines are not processed
            parser.fIn.readline()
            parser.fIn.readline()
            parser.fIn.readline()
            map_index_to_kind = {}

            while match:
                line = parser.fIn.readline()
                result = regex_compiled.match(line)

                if result:
                    match = True
                    kind_number = int(result.groups()[1])
                    element_number = int(result.groups()[3])
                    index = int(result.groups()[0])
                    map_index_to_kind[index] = kind_number
                    info = self.atom_kind_info[kind_number]
                    label = info["label"]
                    info["element_number"] = element_number
                    self.basis_set_info[kind_number]["element_number"] = element_number
                    labels.append(label)
                    coordinate = [float(x) for x in result.groups()[4:]]
                    coordinates.append(coordinate)
                else:
                    match = False
            coordinates = np.array(coordinates)
            labels = np.array(labels)
            self.cache_service["map_index_to_kind"] = map_index_to_kind

            # If anything found, push the results to the correct section
            if len(coordinates) != 0:
                self.cache_service["atom_positions"] = coordinates
                self.cache_service["atom_labels"] = labels

        return wrapper

    def adHoc_run_dir(self, metaname):
        def wrapper(parser):
            end_str = "\n"
            end = False
            path_array = []

            # Loop through lines until empty line is encountered
            while not end:
                line = parser.fIn.readline()
                if line == end_str or len(line) == 0:
                    end = True
                else:
                    path_part = line.split()[-1]
                    path_array.append(path_part)

            # Form the final path and push to backend
            path = "".join(path_array)
            parser.backend.addValue(metaname, path)

        return wrapper

    def adHoc_dft_plus_u(self, parser):
        self.test_electronic_structure_method = "DFT+U"

    def adHoc_mp2(self, parser):
        self.test_electronic_structure_method = "MP2"

    def adHoc_rpa(self, parser):
        self.test_electronic_structure_method = "RPA"

    def adHoc_print(self, msg):
        def wrapper(parser, groups):
            print(msg)
        return wrapper

    #=======================================================================
    # StarReActions
    def startReAction_save_cube_filename(self, backend, groups):
        filename = groups[0]
        self.cache_service["electron_density_filename"] = filename



# Copyright 2019  Alvin Noe Ladines, Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
.. _install-client:

Install the NOMAD client library
________________________________

The NOMAD client library is a Python module (part of the nomad Python package) that
allows to access the NOMAD archive to retrieve and analyse (large amounts) of NOMAD's
archive data. It allows to use queries to filter for desired entries, bulk download
the required parts of the respective archives, and navigate the results using NOMAD's
metainfo Python API.

To install the NOMAD Python package, you can use ``pip install`` to install our
source distribution

.. parsed-literal::

    pip install nomad-lab


First example
_____________

.. literalinclude:: ../examples/client.py
    :language: python

This script should yield a result like this:

.. code::

    Number queries entries: 7628
    Number of entries loaded in the last api call: 10
    Bytes loaded in the last api call: 118048
    Bytes loaded from this query: 118048
    Number of downloaded entries: 10
    Number of made api calls: 1

    Cd2O2: energy -11467.827149010665 hartree
    Sr2O2: energy -6551.45699684026 hartree
    Sr2O2: energy -6551.461104765451 hartree
    Be2O2: energy -178.6990610734937 hartree
    Ca2O2: energy -1510.3938165430286 hartree
    Ca2O2: energy -1510.3937761449583 hartree
    Ba2O2: energy -16684.667362890417 hartree
    Mg2O2: energy -548.9736595672932 hartree
    Mg2O2: energy -548.9724185656775 hartree
    Ca2O2: energy -1510.3908614326358 hartree

Let's discuss the different elements here. First, we have a set of imports. The NOMAD source
codes comes with various sub-modules. The `client` module contains everything related
to what is described here; the `metainfo` is the Python interface to NOMAD's common
archive data format and its data type definitions; the `config` module simply contains
configuration values (like the URL to the NOMAD API).

Next, we create an :class:`ArchiveQuery` instance. This object will be responsible for talking
to NOMAD's API for us in a transparent and lazy manner. This means, it will not download
all data right away, but do so when we are actually iterating through the results.

The archive query takes several parameters:

- The ``query`` is a dictionary of search criteria. The query is used to filter all of NOMAD's
  entry down to a set of desired entries. You can use NOMAD's GUI to create queries and
  copy their Python equivalent with the ``<>``-code button on the result list.
- The ``required`` part, allows to specify what parts of the archive should be downloaded.
  Leave it out to download the whole archives. Based on NOMAD's Metainfo (the 'schema' of
  all archives), you can determine what sections to include and which to leave out. Here,
  we are interested in the first run (usually entries only have one run) and the first
  calculation result.
- With the optional ``per_page`` you can determine, how many results are downloaded at
  a time. For bulk downloading many results, we recommend ~100. If you are just interested
  in the first results a lower number might increase performance.
- With the optional ``max``, we limit the maximum amount of entries that are downloaded,
  just to avoid accidentely iterating through a result set of unknown and potentially large
  size.

When you print the archive query object, you will get some basic statistics about the
query and downloaded data.

The archive query object can be treated as a Python list-like. You use indices and ranges
to select results. Here we iterate through a slice and print the calculated energies
from the first calculation of the entries. Each result is a Python object with attributes
governed by the NOMAD Metainfo. Quantities yield numbers, string, or numpy arrays, while
sub-sections return lists of further objects. Here we navigate the sections ``section_run`` and
sub-section ``section_system`` to access the quantity ``energy_total``. This quantity is a
number with an attached unit (Joule), which can be converted to something else (e.g. Hartree).

The create query object keeps all results in memory. Keep this in mind, when you are
accessing a large amount of query results. You should use :func:`ArchiveQuery.clear`
to remove unnecessary results.

The NOMAD Metainfo
__________________

You can imagine the NOMAD Metainfo as a complex schema for hiearchically organized scientific
data. In this sense, the NOMAD Metainfo is a set of data type definitions. These definitions
then govern how the archive for an data entry in NOMAD might look like. You can browse the
hierarchy of definitions in our `Metainfo browser <../metainfo>`_.

Be aware, that the definitions entail everything that an entry could possibly contain, but
not all entries contain all sections and all quantities. What an entry contains depends
on the information that the respective uploaded data contained, what could be extracted,
and of course what was calculated in the first place. To see what the archive of an concrete
entry looks like, you can use the `search interface <../search>`_, select an entry from the
list fo search results, and click on the *Archive* tab.

To *see inside* an archive object in Python, you can use :func:`nomad.metainfo.MSection.m_to_dict`
which is provided by all archive objects. This will convert a (part of an) archive into a
regular, JSON-serializable Python dictionary.

For more details on the metainfo Python interface, consult the `metainfo documentation <metainfo.html>`_.

The ArchiveQuery class
______________________

.. autoclass:: ArchiveQuery

Working with private data
_________________________

Public NOMAD data can be accessed without any authentication; everyone can use our API
without the need for an account or login. However, if you want to work with your own
data that is not yet published, or embargoed data was shared with you, you need to
authenticate before accessing this data. Otherwise, you will simply not find it with
your queries. To authenticate simply provide your NOMAD username and password to the
:class:`ArchiveQuery` constructor.

'''

from typing import Dict, Union, Any, List
import collections.abc
import requests
from urllib.parse import urlparse
from bravado import requests_client as bravado_requests_client
import time
from keycloak import KeycloakOpenID
from io import StringIO
import math
from urllib.parse import urlencode
import multiprocessing
import json

from nomad import config
from nomad import metainfo as mi
from nomad.datamodel import EntryArchive

# TODO this import is necessary to load all metainfo defintions that the parsers are using
from nomad import parsing  # pylint: disable=unused-import


# This is only necessary to path it during test, because the HTTP client interface differs in test
def get_json(response):
    return response.json()


# This is only necessary to path it during test, because the HTTP client interface differs in test
def get_length(response):
    return len(response.content)


class QueryError(Exception):
    pass


class KeycloakAuthenticator(bravado_requests_client.Authenticator):
    def __init__(self, host, user, password, **kwargs):
        super().__init__(host=host)
        self.user = user
        self.password = password
        self.token = None
        self.__oidc = KeycloakOpenID(**kwargs)

    def apply(self, request=None):
        if self.token is None:
            self.token = self.__oidc.token(username=self.user, password=self.password)
            self.token['time'] = time.time()
        elif self.token['expires_in'] < int(time.time()) - self.token['time'] + 10:
            try:
                self.token = self.__oidc.refresh_token(self.token['refresh_token'])
                self.token['time'] = time.time()
            except Exception:
                self.token = self.__oidc.token(username=self.user, password=self.password)
                self.token['time'] = time.time()

        if request:
            request.headers.setdefault('Authorization', 'Bearer %s' % self.token['access_token'])
            return request
        else:
            return dict(Authorization='Bearer %s' % self.token['access_token'])


class ApiStatistics(mi.MSection):

    nentries = mi.Quantity(
        type=int, default=0,
        description='Number queried entries')

    last_response_nentries = mi.Quantity(
        type=int, default=0,
        description='Number of entries loaded in the last api call')

    last_response_data_size = mi.Quantity(
        type=int, unit=mi.units.bytes, default=0,
        description='Bytes loaded in the last api call')

    loaded_data_size = mi.Quantity(
        type=int, unit=mi.units.bytes, default=0,
        description='Bytes loaded from this query')

    loaded_nentries = mi.Quantity(
        type=int, default=0,
        description='Number of downloaded entries')

    napi_calls = mi.Quantity(
        type=int, default=0,
        description='Number of made api calls')

    def __repr__(self):
        out = StringIO()
        for quantity in self.m_def.all_quantities.values():
            out.write('%s: %s\n' % (quantity.description, self.m_get(quantity)))

        return out.getvalue()


class ProcState:
    '''
    A basic pickable data-class that holds the state of one parallel running
    processes that loads archive API data.
    '''
    def __init__(self, archive_query: 'ArchiveQuery'):
        self.url = archive_query.url
        self.request: Dict[str, Any] = dict(
            query={'$and': archive_query.query},
            required=archive_query.required,
            raise_errors=archive_query.raise_errors)
        self.per_page = archive_query.per_page
        self.authentication = archive_query.authentication

        self.upload_ids: List[str] = []
        self.nentries = 0
        self.total = None
        self.after = None
        self.results = None
        self.error: Exception = None
        self.data_size = 0


def _run_proc(proc_state: ProcState) -> ProcState:
    '''
    The main function for a process that retrieves data from the archive API based
    on its state. Will create a new state. Otherwise it is completely stateless.
    '''
    try:
        url = '%s/%s/%s' % (proc_state.url, 'archive', 'query')

        # create the query
        proc_state.request['query']['upload_id'] = proc_state.upload_ids
        aggregation = proc_state.request.setdefault('aggregation', {'per_page': proc_state.per_page})
        if proc_state.after is not None:
            aggregation['after'] = proc_state.after

        # run the query
        response = requests.post(url, headers=proc_state.authentication, json=proc_state.request)
        if response.status_code != 200:
            if response.status_code == 400:
                message = get_json(response).get('message')
                errors = get_json(response).get('errors')
                if message:
                    raise QueryError('%s: %s' % (message, errors))
                raise QueryError('The query is invalid for unknown reasons (400).')
            raise QueryError(
                'The query is invalid for unknown reasons (%d).' % response.status_code)

        # update the state
        proc_state.data_size += get_length(response)
        data = get_json(response)
        proc_state.results = data.get('results', [])
        proc_state.after = data['aggregation'].get('after', None)
        proc_state.total = data['aggregation'].get('total', 0)
    except Exception as e:
        proc_state.error = e

    return proc_state


class ArchiveQuery(collections.abc.Sequence):
    '''
    Object of this class represent a query on the NOMAD Archive. It is solely configured
    through its constructor. After creation, it implements the
    Python ``Sequence`` interface and therefore acts as a sequence of query results.

    Not all results are downloaded at once, expect that this class will continuesly pull
    results from the API, while you access or iterate to the far side of the result list.

    Attributes:
        query: A dictionary of search parameters. Consult the search API to get a
            comprehensive list of parameters.
        required: A potentially nested dictionary of sections to retrieve.
        url: Optional, override the default NOMAD API url.
        username: Optional, allows authenticated access.
        password: Optional, allows authenticated access.
        per_page: Determine how many results are downloaded per page (or scroll window).
            Default is 10.
        max: Optionally determine the maximum amount of downloaded archives. The iteration
            will stop if max is surpassed even if more results are available. Default is 10.000.
            None value will set it to unlimited.
        raise_errors: There situations where archives for certain entries are unavailable.
            If set to True, this cases will raise an Exception. Otherwise, the entries
            with missing archives are simply skipped (default).
        authentication: Optionally provide detailed authentication information. Usually,
            providing ``username`` and ``password`` should suffice.
        parallel: Number of processes to use to retrieve data in parallel. Only data
            from different uploads can be retrieved in parallel. Default is 1. The
            argument ``per_page`` will refer to archived retrieved in one process per
            call.
    '''
    def __init__(
            self,
            query: dict = None, required: dict = None,
            url: str = None, username: str = None, password: str = None,
            parallel: int = 1, per_page: int = 10, max: int = 10000,
            raise_errors: bool = False,
            authentication: Union[Dict[str, str], KeycloakAuthenticator] = None):

        self.page = 1
        self.parallel = parallel
        self.per_page = per_page
        self.max = max

        self.query: List[dict] = []
        if query is not None:
            self.query.append(query)

        self.raise_errors = raise_errors
        self.required = required if required is not None else dict(section_run='*')
        if required is not None and isinstance(required, dict):
            # We try to add all required properties to the query to ensure that only
            # results with those properties are returned.
            quantities = set()
            required_specs = [required]
            while len(required_specs) > 0:
                for key, value in required_specs.pop().items():
                    section_name = key.split('[')[0]
                    quantities.add(section_name)
                    if isinstance(value, dict):
                        required_specs.append(value)

            self.query.append({'dft.quantities': list(quantities)})
            if 'domain' not in self.query:
                self.query.append({'domain': 'dft' if 'section_experiment' not in quantities else 'ems'})

        self.password = password
        self.username = username
        self.url = config.client.url if url is None else url
        self._authentication = authentication

        self._total = -1
        self._results: List[dict] = []
        self._statistics = ApiStatistics()
        self._proc_states: List[ProcState] = None

    @property
    def authentication(self):
        '''
        The authentication information that is used, if username or password were
        provided.
        '''
        if self._authentication is None and self.username is not None and self.password is not None:
            host = urlparse(self.url).netloc.split(':')[0]
            self._authentication = KeycloakAuthenticator(
                host=host,
                user=self.username,
                password=self.password,
                server_url=config.keycloak.server_url,
                realm_name=config.keycloak.realm_name,
                client_id=config.keycloak.client_id)

        if isinstance(self._authentication, KeycloakAuthenticator):
            return self._authentication.apply()

        else:
            return self._authentication

    def _create_initial_proc_state(self):
        '''
        Does preliminary queries to the repo API to determine the distribution of queried
        entries over uploads and creates initial state for the processes that collect
        data from the archive API in parallel.
        '''
        uploads: Dict[str, Any] = dict()
        nentries = 0

        # acquire all uploads and how many entries they contain
        url = '%s/repo/quantity/upload_id?%s' % (
            self.url, urlencode(dict(query=json.dumps({'$and': self.query}))))
        after: str = None

        while True:
            response = requests.get(
                url if after is None else '%s&after=%s&size=1000' % (url, after),
                headers=self.authentication)

            if response.status_code != 200:
                if response.status_code == 400:
                    raise Exception(response.json()['description'])

                raise Exception(
                    'Error requesting NOMAD API: HTTP %d' % response.status_code)

            response_data = get_json(response)
            after = response_data['quantity']['after']
            values = response_data['quantity']['values']

            if len(values) == 0:
                break

            uploads.update(values)
            for upload in values.values():
                nentries += upload['total']

            if self.max is not None and nentries >= self.max:
                break

        # distribute uploads to processes
        if self.parallel is None:
            self.parallel = 1

        # TODO This implements a simplified distribution, where an upload is fully
        # handled by an individual process. This works because of the high likely hood
        # that popular analysis queries (e.g. AFLOW) have results spread over
        # many uploads. In other use-cases, e.g. analysing data from an individual user,
        # this might not work well, because all entries might be contained in one
        # upload.
        self._proc_states = []
        nentries_per_proc = math.ceil(nentries / self.parallel)
        proc_state = ProcState(self)
        for upload_id, upload_data in uploads.items():
            if proc_state.nentries >= nentries_per_proc:
                self._proc_states.append(proc_state)
                proc_state = ProcState(self)

            proc_state.upload_ids.append(upload_id)
            proc_state.nentries += upload_data['total']

        self._proc_states.append(proc_state)
        self._total = nentries
        self._statistics.nentries = nentries

    def call_api(self):
        '''
        Calls the API to retrieve the next set of results. Is automatically called, if
        not yet downloaded entries are accessed.
        '''
        if self._proc_states is None:
            self._create_initial_proc_state()

        # run the necessary processes
        nproc_states = len(self._proc_states)
        if nproc_states == 1:
            self._proc_states[0] = _run_proc(self._proc_states[0])
        elif nproc_states > 1:
            with multiprocessing.Pool(nproc_states) as pool:
                self._proc_states = pool.map(_run_proc, self._proc_states)
        else:
            assert False, 'archive query was not stopped before running out of things to query'

        # grab the results from the processes
        new_states: List[ProcState] = []
        self._statistics.last_response_nentries = 0
        self._statistics.last_response_data_size = 0
        for proc_state in self._proc_states:
            if proc_state.error:
                raise proc_state.error

            self._statistics.last_response_data_size += proc_state.data_size
            self._statistics.loaded_data_size += proc_state.data_size
            self._statistics.last_response_nentries += len(proc_state.results)

            self._results.extend([
                EntryArchive.m_from_dict(result['archive'])
                for result in proc_state.results])

            proc_state.results = None
            if proc_state.after is not None:
                new_states.append(proc_state)

        self._proc_states = new_states
        self._statistics.loaded_nentries = len(self._results)
        self._statistics.napi_calls += 1

        if self.max is not None and len(self._results) >= self.max:
            # artificially end the query
            self._proc_states = []

        if len(self._proc_states) == 0:
            self._total = len(self._results)

    def __repr__(self):
        if self._total == -1:
            self.call_api()

        return str(self._statistics)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return [self[i] for i in range(*key.indices(len(self)))]

        if key >= self.__len__():
            raise IndexError()

        while len(self._results) < key + 1:
            self.call_api()

        return self._results[key]

    def __len__(self):  # pylint: disable=invalid-length-returned
        if self._total == -1:
            self.call_api()

        return self._total

    @property
    def total(self):
        ''' The total ammount of search results. '''
        if self._total == -1:
            self.call_api()

        return self._total

    @property
    def statistics(self):
        ''' A metainfo object with a basic set of query statistics. '''
        if self._total == -1:
            self.call_api()

        return self._statistics

    def clear(self, index: int = None):
        '''
        Remove caches results. The results are replaced with None in this object. If you
        keep references to the results elsewhere, the garbage collection might not catch
        those.

        Arguments:
            index: Remove all results upto and including the giving index. Default is to
                remove all results.
        '''
        for i, _ in enumerate(self._results[:index]):
            print(i)
            self._results[i] = None


def query_archive(*args, **kwargs):
    return ArchiveQuery(*args, **kwargs)


if __name__ == '__main__':
    run = query_archive()[1]
    run.section_system[1].atom_labels



# Copyright 2016-2018 Carl Poelking, Fawzi Mohamed, Ankit Kariryaa
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import print_function
from builtins import zip
from builtins import str
from builtins import map
from builtins import range
from builtins import object
import os
import sys
import re
import numpy as np

# TODO Distinguish: number_of_steps vs number_of_steps_equilibration
# TODO Check units and install conversions
# TODO - Pressure
# TODO Check sampling_method

UNITCONV_DLPOLY_TO_SI = {
'time' : 1e-12, # ps to s
'length' : 1e-10, # AA to m
'mass' : 1.6605402e-27, # u to kg
'charge' : 1.6021733e-19, # e to C
'energy' : 1.6605402e-23, # 10*J/mol to J
'pressure' : 1.6605402e+7, # 163.88 atm to Pa
'pressure_katm' : 1.01325000e+8, # katm to Pa
'temperature' : 1. # K to K
}

UNITCONV_DLPOLY_CUSTOM_TO_SI = {
'energy_kj_mol' : 1.6605402e-21, # kJ/mol to J
'energy_kj'    : 1.6605402e-21, # kJ/mol to J
'energy_ev' : 1.6021733e-19, # eV to J
'energy_kcal_mol' : 4.184*1.6605402e-21, # kcal/mol to J
'energy_kcal'    : 4.184*1.6605402e-21, # kcal/mol to J
'energy_kelvin_boltzmann' : None,
'energy_k'               : None,
'energy_dl_polyinternalunits_10j_mol' : 1.6605402e-23 # 10*J/mol to J
}

# =================
# TRANSLATION RULES
# =================

KEY_TRANSFORM_SIM_CTRL = {
'simulation_temperature_k' : 'x_dl_poly_thermostat_target_temperature',
'simulation_pressure_katms' : 'x_dl_poly_barostat_target_pressure',
'integration' : 'x_dl_poly_integrator_type',
'ensemble' : 'ensemble_type',
'thermostat_relaxation_time_ps' : 'x_dl_poly_thermostat_tau',
'barostat_relaxation_time_ps' : 'x_dl_poly_barostat_tau',
'selected_number_of_timesteps' : 'x_dl_poly_number_of_steps_requested',
'equilibration_period_steps' : 'number_of_steps_equil_requested',
'temperature_scaling_on_during' : None,
'temperature_scaling_interval' : None,
'equilibration_included_in_overall' : None,
'data_printing_interval_steps' : None,
'statistics_file_interval' : None,
'data_stacking_interval_steps' : None,
'real_space_cutoff_angs' : 'rc_angs',
'vdw_cutoff_angs' : 'rc_vdw_angs',
'electrostatics' : None,
'ewald_sum_precision' : None,
'ewald_convergence_parameter_a^-1' : None,
'extended_coulombic_exclusion' : None,
'cutoff_padding_reset_to_angs' : None,
'vdw_cutoff_reset_to_angs' : None,
'fixed_simulation_timestep_ps' : 'x_dl_poly_integrator_dt',
'data_dumping_interval_steps' : None,
'allocated_job_run_time_s' : None,
'allocated_job_close_time_s' : None
}

KEY_TRANSFORM_SYS_SPEC = {
'energy_units' : 'energy_units',
'number_of_molecular_types' : 'n_molecular_types'
}

KEY_TRANSFORM_MOL_GLOBAL = {
'molecular_species_type' : 'molecule_type_id',
'name_of_species' : 'molecule_type_name',
'number_of_molecules' : 'number_of_molecules'
}

KEY_RULES_CONTROLS = {
'ensemble' :    lambda l: [ s.lower() for s in l[1:] ],
'temperature' : lambda l: l[-1],
'pressure' :    lambda l: l[-1],
'timestep' :    lambda l: l[-1],
'steps':        lambda l: l[-1],
'equilibration':lambda l: l[-1],
'cutoff':       lambda l: l[-1]
}

KEY_RULES_CONTROLS_EXPAND_KEY = {
'ensemble':'ensemble',
'temp':'temperature',
'temperature':'temperature',
'pres':'pressure',
'pressure':'pressure',
'timestep':'timestep',
'steps':'steps',
'equilibration':'equilibration',
'equil':'equilibration',
'cut':'cutoff',
'rcut':'cutoff',
'cutoff':'cutoff'
}


# ===================
# FILE & BLOCK STREAM
# ===================

class FileStream(object):
    def __init__(self, filename=None):
        if filename:
            self.ifs = open(filename, 'r')
        else:
            self.ifs = None
        return
    def SkipTo(self, expr):
        while True:
            ln = self.ifs.readline()
            if expr in ln:
                break
            if self.all_read():
                break
        return ln
    def SkipToMatch(self, expr):
        while True:            
            ln = self.ifs.readline()
            m = re.search(expr, ln)
            if m:
                return ln
            if self.all_read(): break
        return None
    def GetBlock(self, expr1, expr2):
        inside = False
        outside = False
        block = ''
        block_stream = BlockStream()
        while True:
            last_pos = self.ifs.tell()
            ln = self.ifs.readline()
            if expr1 in ln: inside = True
            if expr2 in ln: outside = True
            if inside and not outside:
                # Inside the block
                block += ln
                block_stream.append(ln)
            elif inside and outside:
                self.ifs.seek(last_pos)
                # Block finished
                break
            else:
                # Block not started yet
                pass
            if self.all_read(): break
        return block_stream  
    def GetBlockSequence(self, 
            expr_start, 
            expr_new, 
            expr_end, 
            remove_eol=True, 
            skip_empty=True):
        inside = False
        outside = False
        # Setup dictionary to collect blocks
        blocks = { expr_start : [] }
        for e in expr_new:
            blocks[e] = []
        # Assume structure like this (i <> inside, o <> outside)
        # Lines with 'i' get "eaten"
        # """
        # o ...
        # i <expr_start>
        # i ...
        # i <expr_new[1]>
        # i ...
        # i <expr_new[0]>
        # i ...
        # o <expr_end>
        # o ...
        # """
        key = None
        while True:
            # Log line position
            last_pos = self.ifs.tell()
            ln = self.ifs.readline()            
            # Figure out where we are
            if not inside and expr_start in ln:
                #print "Enter", expr_start
                inside = True
                key = expr_start
                new_block = BlockStream(key)
                blocks[key].append(new_block)
            for expr in expr_new:
                if inside and expr in ln:
                    #print "Enter", expr
                    key = expr
                    new_block = BlockStream(key)
                    blocks[key].append(new_block)
            if inside and expr_end != None and expr_end in ln:
                outside = True
            if inside and not outside:
                # Inside a block
                if remove_eol: ln = ln.replace('\n', '')
                if skip_empty and ln == '': pass
                else: blocks[key][-1].append(ln)
            elif inside and outside:
                # All blocks finished
                self.ifs.seek(last_pos)
                break
            else:
                # No blocks started yet
                pass
            if self.all_read(): break
        return blocks
    def all_read(self):
        return self.ifs.tell() == os.fstat(self.ifs.fileno()).st_size
    def readline(self):
        return ifs.readline()
    def close(self):
        self.ifs.close()
    def nextline(self):
        while True:
            ln = self.ifs.readline().replace('\n','').strip()
            if ln != '':
                return ln
            else: pass
            if self.all_read(): break
        return ln
    def ln(self):
        return self.nextline()
    def sp(self):
        return self.ln().split()
    def skip(self, n):
        for i in range(n):
            self.ln()
        return
    
class BlockStream(FileStream):
    def __init__(self, label=None):
        super(BlockStream, self).__init__(None)
        self.ifs = self
        self.lns = []
        self.idx = 0
        self.label = label
    def append(self, ln):
        self.lns.append(ln)
    def readline(self):
        if self.all_read():
            return ''        
        ln = self.lns[self.idx]
        self.idx += 1
        return ln
    def all_read(self):
        return self.idx > len(self.lns)-1
    def tell(self):
        return self.idx
    def cat(self, remove_eol=True, add_eol=False):
        cat = ''
        for ln in self.lns:
            if remove_eol:
                cat += ln.replace('\n', '')
            elif add_eol:
                cat += ln+'\n'
            else:
                cat += ln
        return cat

# ================
# DLPOLY TERMINALS
# ================

class DlPolyParser(object):
    def __init__(self, log=None):
        self.output_file = 'OUTPUT'
        self.log = log
        self.data = {}
        self.topology = None
        self.logtag = 'sim'
        # KEY DEFAULT DICTIONARIES
        self.missing_keys_lh = [] # Transform keys that were not found in output
        self.missing_keys_rh = []
        self.ignored_keys = [] # Raw keys that did not have a transform
        self.keys_not_found = [] # Searches that failed
        # STORES TIME/STEP VARIABLES (ENERGY, ...)
        self.step_data = {}
        self.time_data = {}
        return
    def __getitem__(self, key):
        self.selected_data_item = self.data[key]
        return self
    def As(self, typ=None):
        if typ == None:
            typ = type(self.selected_data_item)
        return typ(self.selected_data_item)
    def SummarizeKeyDefaults(self):
        if not self.log: return
        if len(self.missing_keys_lh):
            self.log << self.log.my \
                << "[%s] Keys from transformation maps that went unused (=> set to 'None'):" \
                % self.logtag << self.log.endl
            for lh, rh in zip(self.missing_keys_lh, self.missing_keys_rh):
                self.log << self.log.item << "Key = %-25s <> %25s" % (rh, lh) << self.log.endl
        if len(self.ignored_keys):
            self.log << self.log.mb \
                << "[%s] Keys from XY mapping that were not transformed (=> not stored):" \
                % self.logtag << self.log.endl
            for key in self.ignored_keys:
                self.log << self.log.item << "Key =" << key << self.log.endl
        if len(self.keys_not_found):
            self.log << self.log.mr \
                << "[%s] Keys from searches that failed (=> set to 'None'):" \
                % self.logtag << self.log.endl
            for key in self.keys_not_found:
                self.log << self.log.item << "Key =" << key << self.log.endl
        return
    def Set(self, key, value):
        if self.log:
            self.log << "Set [%s]   %-40s = %s" % (self.logtag, key, str(value)) << self.log.endl
        if key not in self.data:
            self.data[key] = value
        else:
            raise KeyError("Key already exists: '%s'" % key)
        return
    def SearchMapKeys(self, expr, ln, keys):
        s = re.search(expr, ln)
        try:
            for i in range(len(keys)):
                self.Set(keys[i], s.group(i+1).strip())
        except AttributeError:
            for i in range(len(keys)):
                self.Set(keys[i], None)
                self.keys_not_found.append(keys[i])
        return
    def ReadBlockXy(self, block):
        lns = block.lns
        block_data = {}
        for ln in lns:
            ln = ln.replace('\n','')
            if ln == '':
                continue
            if ':' in ln:
                sp = ln.split(':')
                x = sp[0].strip().split()
                y = sp[1].strip()
            elif '=' in ln:
                sp = ln.split('=')
                x = sp[0].strip().split()
                y = sp[1].strip()
            else:
                sp = ln.split()
                x = sp[:-1]
                y = sp[-1]
            key = ''
            for i in range(len(x)-1):                
                xi = x[i].replace('(','').replace(')','').lower()
                key += '%s_' % xi
            key += '%s' % x[-1].replace('(','').replace(')','').lower()
            value = y
            block_data[key] = value
        return block_data
    def ApplyBlockXyData(self, block_data, key_map):
        for key_in in key_map:
            key_out = key_map[key_in]            
            if key_in not in block_data:
                # Missing key in output
                self.missing_keys_lh.append(key_in)
                self.missing_keys_rh.append(key_out)
                value = None
            else:
                value = block_data[key_in]
            if key_out == None:
                key_out = key_in
            self.Set(key_out, value)
        for key in block_data:
            if key not in key_map:
                # Missing key in transform map
                self.ignored_keys.append(key)
        return
    def ParseOutput(self, output_file):        
        if self.log: 
            self.log << self.log.mg << "Start simulation method ..." << self.log.endl
        
        ifs = FileStream(output_file)
        self.Set('program_name', 'DL_POLY')
        self.Set('sampling_method', 'molecular_dynamics')
        
        # HEADER & NODE STRUCTURE
        ln = ifs.SkipTo('** DL_POLY **')
        self.SearchMapKeys('version:\s*(\d+.\d+)\s*/\s*(\w+\s*\d+)', ifs.ln(), ['program_version', 'program_version_date'])
        self.SearchMapKeys('execution on\s*(\d+)\s*node', ifs.ln(), ['n_nodes'])
        ln = ifs.SkipTo('node/domain decomposition')
        self.Set('domain_decomposition', list(map(int, ln.split()[-3:])))
        
        # SIMULATION TITLE
        ln = ifs.SkipToMatch('^\s+\*+$')
        ifs.skip(2)
        self.SearchMapKeys('\*+\s([\w+\s\(\)]*)\s\*+', ifs.ln(), ['title'])
        
        # SIMULATION CONTROL PARAMETERS
        block = ifs.GetBlock('SIMULATION CONTROL PARAMETERS', 'SYSTEM SPECIFICATION')        
        block_data = self.ReadBlockXy(block)
        self.ApplyBlockXyData(block_data, KEY_TRANSFORM_SIM_CTRL)        
        
        # TOPOLOGY
        expr_start = 'SYSTEM SPECIFICATION'
        expr_molecule = 'molecular species type'
        expr_config = 'configuration file name'
        expr_vdw = 'number of specified vdw potentials'
        expr_total = 'total number of molecules'
        expr_new = [ expr_molecule, expr_vdw, expr_config, expr_total ]
        expr_end = 'all reading and connectivity checks DONE'
        blocks = ifs.GetBlockSequence(expr_start, expr_new, expr_end)
        # Sanity checks ...
        assert len(blocks[expr_vdw]) == len(blocks[expr_start]) == len(blocks[expr_config]) == 1
        assert len(blocks[expr_molecule]) >= 1
        assert len(blocks[expr_total]) == 1
        block_sys_spec = blocks[expr_start][0]
        block_config = blocks[expr_config][0]
        block_molecules = blocks[expr_molecule]
        block_vdw = blocks[expr_vdw][0]
        # Generate ...  
        self.topology = DlPolyTopology(block_sys_spec, block_config, block_molecules, block_vdw, self)        

        # RUNTIME SYSTEM PROPS: ENERGIES, PRESSURE, ...
        block = ifs.GetBlock('step     eng_tot', 'elapsed cpu time')
        step_start_new = False
        have_step = False
        have_time = False
        step_dict_list = {}
        time_dict_list = {}
        while not block.all_read():
            ln = block.ln()
            if '---------------' in ln:
                step_start_new = True
                continue
            if 'step' in ln:
                block.ln()
                block.ln()
                continue
            sp = ln.split()
            if sp == []: continue
            if step_start_new:
                # step eng_tot ...
                step_dict = {}
                step_nr = int(sp[0])
                step_dict['eng_tot'] = float(sp[1])
                step_dict_list[step_nr] = step_dict
                step_start_new = False
                # time ...
                ln = block.ln()
                sp = ln.split()
                time = float(sp[0])
                time_dict = {}
                time_dict['eng_tot'] = step_dict['eng_tot']
                time_dict_list[time] = time_dict
            else:
                pass
        self.step_data = step_dict_list
        self.time_data = time_dict_list
        #for item in step_dict_list:
        #    print("STEP", item, step_dict_list[item])
        #for item in time_dict_list:
        #    print("TIME", item, time_dict_list[item])
        ifs.close()
        return


class DlPolyControls(DlPolyParser):
    def __init__(self, log=None):
        super(DlPolyControls, self).__init__(log)
        self.logtag = 'ctr'
        return
    def ParseControls(self, ctrl_file):
        if self.log: 
            self.log << self.log.mg << "Start controls ..." << self.log.endl        
        ifs = FileStream(ctrl_file)
        while not ifs.all_read():
            ln = ifs.ln()
            if ln == '' or ln[0:1] == '#': continue
            sp = ln.split()
            key = sp[0]
            try:
                key_long = KEY_RULES_CONTROLS_EXPAND_KEY[key]
                self.Set(key_long, KEY_RULES_CONTROLS[key_long](sp))
            except KeyError:
                self.ignored_keys.append(key)
                pass
        ifs.close()
        return
        

class DlPolyFrame(DlPolyParser):
    def __init__(self, log=None):
        super(DlPolyFrame, self).__init__(log)
        self.atoms = []
        self.has_velocities = False
        self.has_forces = False
        self.has_energy_total = False
        self.energy_total = None
        self.position_matrix = None
        self.pbc_booleans = None
        self.box_matrix = np.zeros((3,3))
    def ParseAtoms(self, ifs, n_atoms, log_level):
        # Log level
        if log_level > 0:
            self.has_velocities = True
            if log_level > 1:
                self.has_forces = True
        # Create atoms
        for i in range(n_atoms):
            atom_name, atom_id = tuple(ifs.ln().split()[0:2]) # This leaves out weight, charge, rsd
            xyz = [float(x) for x in ifs.ln().split()]
            records = [atom_name, atom_id, xyz]
            record_labels = ['atom_name', 'atom_id', 'xyz']
            if log_level > 0:
                vel = [float(x) for x in ifs.ln().split()]
                records.append(vel)
                record_labels.append('vel')
                if log_level > 1:
                    force = [float(x) for x in ifs.ln().split()]
                    records.append(force)
                    record_labels.append('force')
            new_atom = DlPolyAtom(records, record_labels, self)
            self.atoms.append(new_atom)
        assert len(self.atoms) == n_atoms 
        # Position matrix
        self.position_matrix = np.zeros((n_atoms, 3))
        for i in range(n_atoms):
            self.position_matrix[i] = np.array(self.atoms[i]['xyz'].As())
        # Velocity matrix
        if self.has_velocities:
            self.velocity_matrix = np.zeros((n_atoms, 3))
            for i in range(n_atoms):
                self.velocity_matrix[i] = np.array(self.atoms[i]['vel'].As())
        if self.has_forces:
            self.force_matrix = np.zeros((n_atoms, 3))
            for i in range(n_atoms):
                self.force_matrix[i] = np.array(self.atoms[i]['force'].As())
        return
    def ParseBox(self, ifs, pbc_type):
        # PBC booleans
        if pbc_type == 6:
            self.pbc_booleans = np.array([ True, True, False ])
        elif pbc_type > 0:
            self.pbc_booleans = np.array([ True, True, True ])
        else:
            self.pbc_booleans = np.array([ False, False, False ])
        # Box
        if pbc_type > 0:
            a = [float(x) for x in ifs.ln().split()]
            b = [float(x) for x in ifs.ln().split()]
            c = [float(x) for x in ifs.ln().split()]
            self.Set('box_a', a)
            self.Set('box_b', b)
            self.Set('box_c', c)
            # Box matrix
            # ATTENTION First index: x,y,z; second index: a,b,c
            for i in range(3):
                self.box_matrix[i][0] = a[i]
                self.box_matrix[i][1] = b[i]
                self.box_matrix[i][2] = c[i]  
        return
    def ParseFrame(self, ifs, step_data, time_data):
        ln = ifs.ln()
        directives = ln.split()
        #print(directives)
        assert 'timestep' == directives[0]
        # Frame meta-info
        self.Set('timestep', directives[1])
        self.Set('n_atoms', directives[2])
        self.Set('log_level', directives[3])
        self.Set('pbc_type', directives[4])
        self.Set('dt', directives[5])
        self.Set('time_value', self['timestep'].As(int)*self['dt'].As(float))
        n_atoms = self['n_atoms'].As(int)
        pbc_type = self['pbc_type'].As(int)
        # Look-up trajectory data
        step = self['timestep'].As(int)
        if step in step_data:
            if 'eng_tot' in step_data[step]:
                self.has_energy_total = True
                self.energy_total = step_data[step]['eng_tot']
        # Logging level
        log_level = self['log_level'].As(int)
        # Box
        self.ParseBox(ifs, pbc_type)
        # Atoms
        self.ParseAtoms(ifs, n_atoms, log_level)
        return
        

class DlPolyTrajectory(DlPolyParser):
    def __init__(self, log=None):
        super(DlPolyTrajectory, self).__init__(log)
        self.logtag = 'trj'
        self.frames = [] # List of DlPolyFrame's        
    def ParseTrajectory(self, trj_file, step_data, time_data):
        if self.log:
            self.log << self.log.mg << "Start trajectory ..." << self.log.endl
        if not os.path.isfile(trj_file):
            trj_file = '__nofile__'
            pass
        else:
            ifs = FileStream(trj_file)
            title = ifs.ln().replace('\n','').strip()
            # Title
            self.Set('title', title)
            # Directives: logging, pbc
            directives = ifs.ln().split()
            self.Set('log_level', directives[0])
            self.Set('pbc_type', directives[1])
            # Frames
            while not ifs.all_read():
                new_frame = DlPolyFrame(self.log)
                new_frame.ParseFrame(ifs, step_data, time_data)
                self.frames.append(new_frame)
        if self.log:
            self.log << "Read %d frames from %s" % (len(self.frames), trj_file) << self.log.endl
        return


class DlPolyConfig(DlPolyFrame):
    def __init__(self, log=None):
        DlPolyFrame.__init__(self, log)        
        #super(DlPolyConfig, self).__init__(log)
        self.logtag = 'cfg'
        return
    def ParseConfig(self, trj_file):
        if self.log:
            self.log << self.log.mg << "Start configuration ..." << self.log.endl
        ifs = FileStream(trj_file)
        # Title
        title = ifs.ln().replace('\n','').strip()
        self.Set('title', title)
        self.Set('time_value', 0.)
        # Directives: logging, pbc
        directives = ifs.ln().split()
        self.Set('log_level', directives[0]) # 0 -> 1 -> 2: coords -> + vel. -> + forces
        self.Set('pbc_type', directives[1])  # 0 / ... / 6: no / cubic / orthorhom. / par.-epiped / xy
        self.Set('n_atoms', directives[2])
        n_atoms = self['n_atoms'].As(int)
        pbc_type = self['pbc_type'].As(int)
        # Logging level
        log_level = self['log_level'].As(int)
        # Box
        self.ParseBox(ifs, pbc_type)
        # Atom records
        self.ParseAtoms(ifs, n_atoms, log_level)
        return

# =======================
# DLPOLY TOPOLOGY OBJECTS
# =======================

class DlPolyTopology(DlPolyParser):
    def __init__(self, block_sys_spec, block_config, block_mols, block_vdw, parser):
        super(DlPolyTopology, self).__init__(parser.log)
        self.logtag = 'top'
        
        if self.log: self.log << self.log.mg << "Start topology ..." << self.log.endl
        
        # Meta specification (energy values, # molecular types)
        sys_spec = parser.ReadBlockXy(block_sys_spec)
        self.ApplyBlockXyData(sys_spec, KEY_TRANSFORM_SYS_SPEC)
        
        # Config specification (config-file name/title, box vectors, box volume)
        config_str = block_config.cat(remove_eol=False, add_eol=True)
        self.SearchMapKeys('configuration file name:\s([\w+\s\(\)]*)\s\n', config_str, ['config_file_name'])
        self.SearchMapKeys('selected image convention\s*(\d+)\s*\n', config_str, ['image_convention'])
        triple_str='\s*([0-9a-zA-Z_.]*)'*3+'\n'
        search_str = 'simulation cell vectors\s*\n' + 3*triple_str
        self.SearchMapKeys(search_str, config_str, ['box_ax', 'box_ay', 'box_az', 'box_bx', 'box_by', 'box_bz', 'box_cx', 'box_cy', 'box_cz'])
        self.SearchMapKeys('system volume\s*([-+0-9.eEdD]*)\s*\n', config_str, ['box_volume'])        
        
        # Molecule specification
        self.molecules = []
        for block_mol in block_mols:
            if self.log: self.log << self.log.mg << "Start molecule ..." << self.log.endl
            new_mol = DlPolyMolecule(block_mol, self)
            self.molecules.append(new_mol)
        
        n_atoms_total = 0
        n_molecules_total = 0
        for mol in self.molecules:
            n_mol = mol['number_of_molecules'].As(int)
            n_atoms = mol['number_of_atoms_in_molecule'].As(int)
            n_atoms_total += n_mol*n_atoms
            n_molecules_total += n_mol
        self.Set('number_of_topology_atoms', n_atoms_total)
        self.Set('number_of_topology_molecules', n_molecules_total)
        return


class DlPolyMolecule(DlPolyParser):
    def __init__(self, mol_stream, parser):
        super(DlPolyMolecule, self).__init__(parser.log)
        self.logtag = 'mol'
        self.atoms = []
        
        # PARTITION ONTO BLOCKS
        expr_global = 'molecular species type'
        # Atoms ...
        expr_atoms = 'number of atoms/sites'
        # Interactions ...
        expr_bonds = 'number of bonds'
        expr_bond_constraints = 'number of bond constraints'
        expr_angles = 'number of bond angles'
        expr_dihedrals = 'number of dihedral angles'
        expr_inv_angles = 'number of inversion angles'
        # Block definitions ...
        expr_start = expr_global
        expr_new = [ expr_atoms, expr_bonds, expr_bond_constraints, expr_angles, expr_dihedrals, expr_inv_angles ]
        expr_end = None        
        blocks = mol_stream.GetBlockSequence(expr_start, expr_new, expr_end)
        # Sanity checks ...
        for key in expr_new:
            assert len(blocks[key]) <= 1
        assert len(blocks[expr_atoms]) == 1
        
        # PARSE GLOBALS
        block = blocks[expr_global][0]
        block_data = self.ReadBlockXy(block)
        self.ApplyBlockXyData(block_data, KEY_TRANSFORM_MOL_GLOBAL)
        
        # PARSE ATOMS
        if self.log: self.log << self.log.mg << "Start atoms ..." << self.log.endl
        block = blocks[expr_atoms][0]
        n_atoms = int(block.ln().split()[-1])
        self.Set('number_of_atoms_in_molecule', n_atoms)
        assert 'atomic characteristics' in block.ln()
        # Determine atom properties        
        atom_property_labels = block.ln().split()
        assert atom_property_labels[0] == 'site'
        atom_property_labels[0] = 'id'
        atom_property_labels = [ 'atom_%s' % l.lower() for l in atom_property_labels ]
        atom_count = 0
        # Read atom lines & create atoms
        while not block.all_read():
            atom_properties = block.ln().split()
            new_atom = DlPolyAtom(atom_properties, atom_property_labels, parser)
            self.atoms.append(new_atom)
            # Atom may repeat - make these repititions explicit
            atom_id = new_atom['atom_id'].As(int)
            atom_repeat = new_atom['atom_repeat'].As(int)
            for i in range(atom_repeat-1):
                next_id = atom_id+i+1
                assert int(atom_properties[0]) == next_id-1
                atom_properties[0] = atom_id+i+1
                new_atom = DlPolyAtom(atom_properties, atom_property_labels, parser)
                self.atoms.append(new_atom)
            # Keep track of total count
            atom_count += atom_repeat
            assert atom_count <= n_atoms
            if atom_count == n_atoms:
                break
        assert atom_count == n_atoms
        
        atom_charges = [ atom['atom_charge'].As(float) for atom in self.atoms ]
        atom_names = [ atom['atom_name'].As() for atom in self.atoms ]
        self.Set('atom_in_molecule_charge', atom_charges)
        self.Set('atom_in_molecule_name', atom_names)
        
        # TODO Parse interactions
        return


class DlPolyAtom(DlPolyParser):
    def __init__(self, atom_properties, atom_property_labels, parser):
        super(DlPolyAtom, self).__init__(parser.log)
        if self.log and not self.log.debug: self.log = None
        self.logtag = 'atm'
        for value, label in zip(atom_properties, atom_property_labels):
            self.Set(label, value)
        return
        
        
        




try:
    from optimade.filtertransformers.django import Lark2Django

    DJANGO_NOT_IMPORTED = False
except ImportError:
    DJANGO_NOT_IMPORTED = True

from unittest import TestCase, skipIf

test_data = [
    ("band_gap<1", "(AND: ('calculation__band_gap__lt', '1'))"),
    (
        "(natoms >= 8) OR (nelements<5 AND stability>=0.5)",
        "(OR: ('entry__natoms__gte', '8'), (AND: ('entry__composition__ntypes__lt', '5'), ('stability__gte', '0.5')))",
    ),
    ("NOT natoms >= 8", "(NOT (AND: ('entry__natoms__gte', '8')))"),
    (
        "element = Th AND element != Na ",
        "(AND: ('composition__element_list__contains', 'Th'), (NOT (AND: ('composition__element_list__contains', 'Na'))))",
    ),
]


@skipIf(DJANGO_NOT_IMPORTED, "Django not found")
class DjangoParserTest(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.Transformer = Lark2Django()

    def test_query_conversion(self):
        for raw_q, dj_q in test_data:
            parsed_tree = self.Transformer.parse_raw_q(raw_q)
            self.assertEqual(str(self.Transformer.evaluate(parsed_tree)), dj_q)



# Copyright 2015-2018 Mikkel Strange, Fawzi Mohamed, Ankit Kariryaa, Ask Hjorth Larsen, Jens Jørgen Mortensen
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import division
import os
from contextlib import contextmanager
import numpy as np
import logging
from ase import units
from ase.data import chemical_symbols
from ase.io.ulm import ulmopen
from ase.utils import basestring
#from ase.io.trajectory import read_atoms
from ase.data import atomic_masses
from ase.units import Rydberg
from nomadcore.unit_conversion.unit_conversion import convert_unit as cu
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from gpawparser.libxc_names import get_libxc_name
from gpawparser.default_parameters import parameters as parms


@contextmanager
def open_section(p, name):
    gid = p.openSection(name)
    yield gid
    p.closeSection(name, gid)


def c(value, unit=None):
    """ Dummy function for unit conversion"""
    return cu(value, unit)


parser_info = {"name": "parser2_gpaw", "version": "1.0"}
path = '../../../../nomad-meta-info/meta_info/nomad_meta_info/' +\
        'gpaw.nomadmetainfo.json'

class GPAWParser2Wrapper():
    """ A proper class envolop for running this parser using Noamd-FAIRD infra. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
        logging.info('GPAW parser 2 (note the 2) started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory("gpaw.nomadmetainfo.json")
        backend = parse_without_class(mainfile, backend)
        return backend


def parse_without_class(filename, backend):
    p = backend
    o = open_section
    r = ulmopen(filename) #  Reader(filename)
    p.startedParsingSession(filename, parser_info)
    parms.update(r.parameters.asdict())
    with o(p, 'section_run'):
        p.addValue('program_name', 'GPAW')
        p.addValue('program_version', r.gpaw_version)
        mode = parms['mode']
        if isinstance(mode, basestring):
            mode = {'name': mode}
        if mode['name'] == 'pw':
            p.addValue('program_basis_set_type', 'plane waves')
            with o(p, 'section_basis_set_cell_dependent'):
                p.addValue('basis_set_cell_dependent_name',
                           'PW_%.1f_Ry' % (mode['ecut'] / r.ha * 2))  # in Ry
                p.addRealValue('basis_set_planewave_cutoff',
                               c(mode['ecut'], 'eV'))
        elif mode['name'] == 'fd':
            p.addValue('program_basis_set_type', 'real space grid')
            with o(p, 'section_basis_set_cell_dependent'):
                cell = r.atoms.cell
                ngpts = r.density.density.shape
                h1 = np.linalg.norm(cell[0]) / ngpts[0]
                h2 = np.linalg.norm(cell[1]) / ngpts[1]
                h3 = np.linalg.norm(cell[2]) / ngpts[2]
                h = (h1 + h2 + h3) / 3.0
                p.addValue('basis_set_cell_dependent_name',
                           'GR_%.1f' % (c(h, 'angstrom') * 1.0E15))  # in fm
        elif mode['name'] == 'lcao':
            p.addValue('program_basis_set_type', 'numeric AOs')
            with o(p, 'section_basis_set_atom_centered'):
                p.addValue('basis_set_atom_centered_short_name',
                          parms['basis'])
        with o(p, 'section_system') as system_gid:
            p.addArrayValues('simulation_cell',
                             c(r.atoms.cell, 'angstrom'))
            symbols = np.array([chemical_symbols[z] for z in r.atoms.numbers])
            p.addArrayValues('atom_labels', symbols)
            p.addArrayValues('atom_positions', c(r.atoms.positions, 'angstrom'))
            p.addArrayValues('configuration_periodic_dimensions',
                             np.array(r.atoms.pbc, bool))
            if 'momenta' in r.atoms:
                masses = atomic_masses[r.atoms.numbers]
                velocities = r.atoms.momenta / masses.reshape(-1, 1)
                p.addArrayValues('atom_velocities',
                                c(velocities * units.fs / units.Angstrom,
                                  'angstrom/femtosecond'))
        with o(p, 'section_sampling_method'):
            p.addValue('ensemble_type', 'NVE')
        with o(p, 'section_frame_sequence'):
            pass
        with o(p, 'section_method') as method_gid:
            p.addValue('relativity_method', 'pseudo_scalar_relativistic')
            p.addValue('electronic_structure_method', 'DFT')
            p.addValue('scf_threshold_energy_change',
                       c(parms['convergence']['energy'], 'eV')) # eV / electron
            #if r.FixMagneticMoment:
            #    p.addValue('x_gpaw_fixed_spin_Sz',
            #               r.MagneticMoments.sum() / 2.)
            if parms['occupations'] is None:  # use default values
                if tuple(parms['kpts']) == (1, 1, 1):
                    width = 0.0
                else:
                    width = 0.1
                parms['occupations'] = {'width': width, 'name': 'fermi-dirac'}

            p.addValue('smearing_kind', parms['occupations']['name'])
            p.addRealValue('smearing_width',
                               c(parms['occupations']['width'], 'eV'))
            p.addRealValue('total_charge', parms['charge'])
            with o(p, 'section_XC_functionals'):
                p.addValue('XC_functional_name',
                           get_libxc_name(parms['xc']))
        with o(p, 'section_single_configuration_calculation'):
            p.addValue('single_configuration_calculation_to_system_ref',
                       system_gid)
            p.addValue('single_configuration_to_calculation_method_ref',
                       method_gid)
            p.addValue('single_configuration_calculation_converged',
                      r.scf.converged)
            p.addRealValue('energy_total',
                           c(r.hamiltonian.e_total_extrapolated, 'eV'))
            p.addRealValue('energy_free',
                           c(r.hamiltonian.e_total_free, 'eV'))
            p.addRealValue('energy_XC', c(r.hamiltonian.e_xc, 'eV'))
            p.addRealValue('electronic_kinetic_energy',
                           c(r.hamiltonian.e_kinetic, 'eV'))
            p.addRealValue('energy_correction_entropy',
                           c(r.hamiltonian.e_entropy, 'eV'))
            nspin = r.density.density.shape[0]
            ef = r.occupations.fermilevel
            fermilevels = np.zeros(nspin)
            if nspin == 1:
                fermilevels[:] = ef
            elif nspin == 2:
                split = r.occupations.split
                fermilevels[0] = ef + 0.5 * split
                fermilevels[1] = ef - 0.5 * split
            p.addArrayValues('energy_reference_fermi',
                             c(fermilevels, 'eV'))

            # Load 3D arrays ("volumetric data")
            origin = -0.5 * r.atoms.cell.sum(axis=0)
            npoints = np.array(r.density.density.shape[1:])
            npoints[~r.atoms.pbc] += 1
            displacements = r.atoms.cell / npoints

            def add_3d_array(values, kind, unit):
                with o(p, 'section_volumetric_data'):
                    p.addArrayValues('volumetric_data_origin',
                                     c(origin, 'angstrom'))
                    p.addArrayValues('volumetric_data_displacements',
                                     c(displacements, 'angstrom'))
                    p.addArrayValues('volumetric_data_values',
                                     c(values, unit))
                    p.addValue('volumetric_data_kind', kind)

            # H.atom.ulm.gpw test can be used to verify that pseudodensity
            # integrates to 0.98, corresponding closely to the norm of the
            # H 1s pseudowavefunction (see output of "gpaw-setup H").
            # It has mixed BCs so this should show that npoints is taken
            # care of correctly, hopefully.
            add_3d_array(r.density.density, kind='density',
                         unit='angstrom**(-3)')
            add_3d_array(r.hamiltonian.potential, kind='potential_effective',
                         unit='eV*angstrom**(-3)')

            if 'forces' in r.results:
                p.addArrayValues('atom_forces_free_raw',
                                 c(r.results.forces, 'eV/angstrom'))
            if 'magmons' in r.results:
                p.addArrayValues('x_gpaw_magnetic_moments',
                                 r.results.magmoms)
                p.addRealValue('x_gpaw_spin_Sz', r.results.magmoms.sum() / 2.0)
            #p.addArrayValues('x_gpaw_atomic_density_matrices',
            #                 r.AtomicDensityMatrices)
            #p.addArrayValues('x_gpaw_projections_real', r.Projections.real)
            #p.addArrayValues('x_gpaw_projections_imag', r.Projections.imag)
            with o(p, 'section_eigenvalues'):
                p.addValue('eigenvalues_kind', 'normal')
                p.addArrayValues('eigenvalues_values',
                                 c(r.wave_functions.eigenvalues, 'eV'))
                p.addArrayValues('eigenvalues_occupation',
                                 r.wave_functions.occupations)
                #p.addArrayValues('eigenvalues_kpoints', r.IBZKPoints)
            if 'band_paths' in r.wave_functions: # could change
                with o(p, 'section_k_band'):
                    for band_path in r.wave_functions.band_paths:
                        with o(p, 'section_k_band_segment'):
                            p.addArrayValues('band_energies',
                                            c(band_path.eigenvalues, 'eV'))
                            p.addArrayValues('band_k_points', 'eV',
                                             band_path.kpoints)
                            p.addArrayValues('band_segm_labels',
                                            band_path.labels)
                            p.addArrayValues('band_segm_start_end',
                                             np.asarray(
                                                 [band_path.kpoints[0],
                                                  band_path.kpoints[-1]]))


    p.finishedParsingSession("ParseSuccess", None)
    return p

if __name__ == '__main__':
    import sys
    filename = sys.argv[1]
    parse(filename)



import logging

# create console handler with a higher log level
# ch = logging.StreamHandler()
# ch.setLevel(logging.INFO)

logger=logging.getLogger("nomadcore")
logger.setLevel(logging.WARNING)
# logger.addHandler(ch)

logger2=logging.getLogger("nomad")
logger2.setLevel(logging.WARNING)
# logger2.addHandler(ch)

def debugToFile():
    "makes a full log to a file named detailed.log"
    fh = logging.FileHandler('detailed.log')
    fh.setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)
    logger2.setLevel(logging.DEBUG)
    logger.addHandler(fh)
    logger2.addHandler(fh)



# Copyright 2015-2018 Martina Stella, Massimo Riello, Fawzi Mohamed, Ankit Kariryaa
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import numpy as np
import math
from nomadcore.simple_parser import mainFunction
from nomadcore.simple_parser import SimpleMatcher as SM
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from nomadcore.caching_backend import CachingLevel
import re, os, sys, json, logging



################################################################################################################################################################
################################################################################################################################################################
################################################################################################################################################################
######################  PARSER CONTEXT CLASS  ##################################################################################################################
################################################################################################################################################################
############################################################################ CASTEP.Parser Version 1.0 #########################################################
################################################################################################################################################################

class CastepParserContext(object):

    def __init__(self):
        self.cell                              = []
        self.at_nr                             = 0
        self.atom_labels                        = []
        self.atom_forces                       = []
        self.castep_atom_positions              = []
        self.atom_positions                     = []
        self.a                                 = []
        self.b                                 = []
        self.c                                 = []
        self.alpha                             = []
        self.beta                              = []
        self.gamma                             = []
        self.volume                            = 0

        self.energy_total_scf_iteration_list   = []
        self.scfIterNr                         = []
        self.ecut                              = []
        self.k_count                           = 0
        self.k_nr                              = 0
        self.e_nr                              = 0
        self.k_count_1                         = 0
        self.k_nr_1                            = 0
        self.e_nr_1                            = 0
        self.castep_band_kpoints               = []
        self.castep_band_energies              = []
        self.castep_band_kpoints_1             = []
        self.castep_band_energies_1            = []



# Translating the XC functional name to the NOMAD standard (as seen in the CP2K parser, thanks!)
    def onClose_castep_section_functionals(self, backend, gIndex, section):
        """When all the functional definitions have been gathered, matches them
        with the nomad correspondents and combines into one single string which
        is put into the backend.
        """
        # Get the list of functional and relativistic names
        functional_names = section["castep_functional_name"]
        relativistic_names = section["castep_relativity_treatment_scf"]

        # Define a mapping for the functionals
        functional_map = {
            " Perdew Burke Ernzerhof": "GGA_C_PBE",
            " Local Density Approximation": "LDA_C_PZ",
            " Perdew Wang (1991)": "GGA_X_PW91",
            " revised Perdew Burke Ernzerhof": "GGA_X_RPBE",
            " PBE with Wu-Cohen exchange": "GGA_X_WC",
            " PBE for solids (2008)": "GGA_X_PBE_SOL",
        }

        # Define a mapping for the relativistic treatments
        relativistic_map = {
            " Koelling-Harmon": "scalar_relativistic"
        }

        # Match each castep functional name and sort the matches into a list
        functionals = []

        for name in functional_names:
            match = functional_map.get(name)
            if match:
                functionals.append(match)
        functionals = "_".join(sorted(functionals))

        # Push the functional string into the backend
        backend.addValue('XC_functional', functionals)

        # Match each castep relativity treatment name and sort the matches into a list
        relativistic = []

        for name in relativistic_names:
            match = relativistic_map.get(name)
            if match:
                relativistic.append(match)
        relativistic = "_".join(sorted(relativistic))

        # Push the relativistic treatment string into the backend
        backend.addValue('relativity_method', relativistic)
        backend.addValue('XC_method_current', functionals+'_'+relativistic)




# Storing the unit cell
    def onClose_castep_section_cell(self, backend, gIndex, section):
        """trigger called when _castep_section_cell is closed"""
        # get cached values for castep_cell_vector
        vet = section['castep_cell_vector']

        vet[0] = vet[0].split()
        vet[0] = [float(i) for i in vet[0]]

        vet[1] = vet[1].split()
        vet[1] = [float(i) for i in vet[1]]

        vet[2] = vet[2].split()
        vet[2] = [float(i) for i in vet[2]]

        self.cell.append(vet[0])
        self.cell.append(vet[1])
        self.cell.append(vet[2]) # Reconstructing the unit cell vector by vector
        backend.addArrayValues('simulation_cell', np.asarray(self.cell), unit='angstrom')



# Storing the total energy of each SCF iteration in an array
    def onClose_section_scf_iteration(self, backend, gIndex, section):
        """trigger called when _section_scf_iteration is closed"""
        # get cached values for energy_total_scf_iteration
        ev = section['energy_total_scf_iteration']
        self.scfIterNr = len(ev)
        self.energy_total_scf_iteration_list.append(ev)

        backend.addArrayValues('energy_total_scf_iteration_list', np.asarray(self.energy_total_scf_iteration_list))
        backend.addValue('number_of_scf_iterations', self.scfIterNr)



# Here we recover the unit cell dimensions (both magnitudes and angles) (useful to convert fractional coordinates to cartesian)
    def onClose_castep_section_atom_positions(self, backend, gIndex, section):
        """trigger called when _castep_section_atom_positions is closed"""
        # get cached values for cell magnitudes and angles
        self.a = section['castep_cell_length_a']
        self.b = section['castep_cell_length_b']
        self.c = section['castep_cell_length_c']
        self.alpha = section['castep_cell_angle_alpha']
        self.beta  = section['castep_cell_angle_beta']
        self.gamma = section['castep_cell_angle_gamma']
        self.volume = np.sqrt( 1 - math.cos(np.deg2rad(self.alpha[0]))**2
                                 - math.cos(np.deg2rad(self.beta[0]))**2
                                 - math.cos(np.deg2rad(self.gamma[0]))**2
                                 + 2 * math.cos(np.deg2rad(self.alpha[0]))
                                     * math.cos(np.deg2rad(self.beta[0]))
                                     * math.cos(np.deg2rad(self.gamma[0])) ) * self.a[0]*self.b[0]*self.c[0]



# Here we add basis set name and kind for the plane wave code
    def onClose_section_basis_set_cell_dependent(self, backend, gIndex, section):
        ecut_str = section['castep_basis_set_planewave_cutoff']
        self.ecut = float(ecut_str[0])
        eVtoRy = 0.073498618
        ecut_str_name = int(round(eVtoRy*self.ecut))

        basis_set_kind = 'plane_waves'
        basis_set_name = 'PW_'+str(ecut_str_name)
        backend.addValue('basis_set_planewave_cutoff', self.ecut)
        backend.addValue('basis_set_cell_dependent_kind', basis_set_kind)
        backend.addValue('basis_set_cell_dependent_name', basis_set_name)



######################################################################################
################ Triggers on closure section_system ######################
######################################################################################

    def onClose_section_system(self, backend, gIndex, section):
        """trigger called when _section_system is closed"""


# Processing forces acting on atoms (final converged forces)
        #get cached values of castep_store_atom_forces
        f_st = section['castep_store_atom_forces']
        self.at_nr = len(f_st)
        for i in range(0, self.at_nr):
            f_st[i] = f_st[i].split()
            f_st[i] = [float(j) for j in f_st[i]]
            f_st_int = f_st[i]
            self.atom_forces.append(f_st_int)
        backend.addArrayValues('atom_forces', np.asarray(self.atom_forces))



# Processing the atom positions in fractionary coordinates (as given in the CASTEP output)
        #get cached values of castep_store_atom_positions
        pos = section['castep_store_atom_positions']
        for i in range(0, self.at_nr):
            pos[i] = pos[i].split()
            pos[i] = [float(j) for j in pos[i]]
            self.castep_atom_positions.append(pos[i])
        backend.addArrayValues('castep_atom_positions', np.asarray(self.castep_atom_positions))



# Processing the atom labels
        #get cached values of castep_store_atom_labels
        lab = section['castep_store_atom_labels']
        for i in range(0, self.at_nr):
            lab[i] = re.sub('\s+', ' ', lab[i]).strip()
        self.atom_labels.append(lab)
        backend.addArrayValues('atom_labels', np.asarray(self.atom_labels))



# Converting the fractional atomic positions (x) to cartesian coordinates (X) ( X = M^-1 x )
        for i in range(0, self.at_nr):

            pos_a = [   self.a[0] * self.castep_atom_positions[i][0]
                      + self.b[0] * math.cos(np.deg2rad(self.gamma[0])) * self.castep_atom_positions[i][1]
                      + self.c[0] * math.cos(np.deg2rad(self.beta[0])) * self.castep_atom_positions[i][2],

                        self.b[0] * math.sin(self.gamma[0]) * self.castep_atom_positions[i][1]
                      + self.c[0] * self.castep_atom_positions[i][2] * (( math.cos(np.deg2rad(self.alpha[0]))
                      - math.cos(np.deg2rad(self.beta[0])) * math.cos(np.deg2rad(self.gamma[0])) ) / math.sin(np.deg2rad(self.gamma[0])) ),

                       (self.volume / (self.a[0]*self.b[0] * math.sin(np.deg2rad(self.gamma[0])))) * self.castep_atom_positions[i][2] ]

            self.atom_positions.append(pos_a)
        backend.addArrayValues('atom_positions', np.asarray(self.atom_positions))



# Backend add the total number of atoms in the simulation cell
        backend.addValue('number_of_atoms', self.at_nr)



######################################################################################
###################### Storing k points and band energies ############################
############################# FIRST SPIN CHANNEL #####################################
######################################################################################

# Storing the k point coordinates (SPIN 1)
    def onClose_castep_section_k_points(self, backend, gIndex, section):
        """trigger called when _section_eigenvalues"""
# Processing k points (given in fractional coordinates)
        #get cached values of castep_store_k_points
        k_st = section['castep_store_k_points']
        self.k_count = len(k_st)
        self.k_nr   += 1
        for i in range(0, self.k_count):
            k_st[i] = k_st[i].split()
            k_st[i] = [float(j) for j in k_st[i]]
            k_st_int = k_st[i]
            self.castep_band_kpoints.append(k_st_int)



# Storing the eigenvalues (SPIN 1)
    def onClose_castep_section_eigenvalues(self, backend, gIndex, section):
        """trigger called when _section_eigenvalues"""
        #get cached values of castep_store_k_points
        e_st = section['castep_store_eigenvalues']
        self.e_nr = len(e_st)
        self.castep_band_energies.append(e_st)



######################################################################################
###################### Storing k points and band energies ############################
############################# SECOND SPIN CHANNEL ####################################
######################################################################################

# Storing the k point coordinates (SPIN 2)
    def onClose_castep_section_k_points_1(self, backend, gIndex, section):
        """trigger called when _section_eigenvalues"""
# Processing k points (given in fractional coordinates)
        #get cached values of castep_store_k_points
        k_st_1 = section['castep_store_k_points_1']
        self.k_count_1 = len(k_st_1)
        self.k_nr_1   += 1
        for i in range(0, self.k_count_1):
            k_st_1[i] = k_st_1[i].split()
            k_st_1[i] = [float(j) for j in k_st_1[i]]
            k_st_1_int = k_st_1[i]
            self.castep_band_kpoints.append(k_st_1_int)

        self.k_nr_1 = self.k_nr  # clean double counting



# Storing the eigenvalues (SPIN 2)
    def onClose_castep_section_eigenvalues_1(self, backend, gIndex, section):
        """trigger called when _section_eigenvalues"""
        #get cached values of castep_store_k_points
        e_st_1 = section['castep_store_eigenvalues_1']
        self.e_nr_1 = len(e_st_1)
        self.castep_band_energies.append(e_st_1)

        self.e_nr_1 = self.e_nr


# Keeping only the last arrays updateclear

    def onClose_castep_section_k_band(self, backend, gIndex, section):
        """trigger called when _section_eigenvalues"""
        backend.addArrayValues('castep_band_kpoints', np.asarray(self.castep_band_kpoints))
        backend.addArrayValues('castep_band_energies', np.asarray(self.castep_band_energies))
# Backend add the number of k points and eigenvalues
        backend.addValue('castep_number_of_band_kpoints', self.k_nr)
        backend.addValue('castep_number_of_band_energies', self.e_nr)

        if np.size(self.castep_band_kpoints) > self.k_nr:  # Clean up spin-1 info prior ejecting spin-2
            self.castep_band_kpoints = []
        if np.size(self.castep_band_energies) > self.k_nr:
            self.castep_band_energies = []



################################################################################################################################################################
################################################################################################################################################################
################################################################################################################################################################
######################  MAIN PARSER STARTS HERE  ###############################################################################################################
################################################################################################################################################################
############################################################################ CASTEP.Parser Version 1.0 #########################################################
################################################################################################################################################################

# main Parser
mainFileDescription = SM(name = 'root',
                         weak = True,
                         startReStr = "",
                         subMatchers = [

                            SM(name = 'newRun',
                            startReStr = r"\s\|\s*CCC\s*AA\s*SSS\s*TTTTT\s*EEEEE\s*PPPP\s*\|\s*",
                            #repeats = True,
                            required = True,
                            forwardMatch = True,
                            sections   = ['section_run'],
                            subMatchers = [

                                SM(name = 'ProgramHeader',
                                   startReStr = r"\s\|\s*CCC\s*AA\s*SSS\s*TTTTT\s*EEEEE\s*PPPP\s*\|\s*",
                                   subMatchers = [

                                    SM(r"\s\|\sWelcome to Academic Release\s(?P<program_name>[a-zA-Z]+)* version *(?P<program_version>[0-9a-zA-Z_.]*)"),
                                    SM(r"\sCompiled for *(?P<program_compilation_host>[-a-zA-Z0-9._]*)\son\s(?P<castep_program_compilation_date>[a-zA-Z,\s0-9]*)\s *(?P<castep_program_compilation_time>[0-9:]*)"),
                                    SM(r"\sCompiler\: *(?P<castep_compiler>[a-zA-Z\s0-9.]*)"),
                                    SM(r"\sMATHLIBS\: *(?P<castep_maths_library>[a-zA-Z0-9.() ]*)\s*"),
                                    SM(r"\sFFT Lib \: *(?P<castep_fft_library>[a-zA-Z0-9.() ]*)\s*"),
                                    SM(r"\sFundamental constants values\: *(?P<castep_constants_reference>[a-zA-Z0-9.() ]*)\s*"),

                                  ]), # CLOSING SM ProgramHeader


                                SM(startReStr = r"\sCalculation not parallelised\.",
                                   sections = ["section_system"],
                                   subMatchers = [

                                       # XC functional
                                       SM(name = 'XCMethods',
                                          startReStr = r"\susing functional\s*\:",
                                          forwardMatch = True,
                                          sections = ["section_method"],
                                          subMatchers = [

                                              SM(name = "castepXC",
                                                 startReStr = r"\susing functional\s*\:",
                                                 forwardMatch = True,
                                                 sections = ["castep_section_functionals"],
                                                 subMatchers = [

                                                    SM(r"\susing functional\s*\: *(?P<castep_functional_name> [A-Za-z0-9() ]*)"),
                                                    SM(r"\srelativistic treatment\s*\: *(?P<castep_relativity_treatment_scf> [A-Za-z0-9() -]*)")

                                                 ]), # CLOSING castep_section_functionals


                                          ]), # CLOSING section_method


                                       # plane wave energy cutoff
                                       SM(name = 'planeWave basis set',
                                          startReStr = r"\sbasis set accuracy\s*",
                                          forwardMatch = True,
                                          sections = ["section_basis_set_cell_dependent"],
                                          subMatchers = [

                                              SM(r"\splane wave basis set cut\-off\s*\:\s*(?P<castep_basis_set_planewave_cutoff>[0-9.]+)")

                                          ]), # CLOSING section_basis_set_cell_dependent


                                       # cell information
                                       SM(name = 'cellInformation',
                                          startReStr = r"\s*Unit Cell\s*",
                                          forwardMatch = True,
                                          sections = ["castep_section_cell"],
                                          subMatchers = [

                                            SM(r"\s*(?P<castep_cell_vector>[\d\.]+\s+[\d\.]+\s+[\d\.]+) \s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*\s*[-+0-9.eEdD]*",
                                               endReStr = "\n",
                                               repeats = True),

                                          ]), # CLOSING castep_section_cell


                                       # atomic positions and cell dimesions
                                       SM(startReStr = r"\s*Lattice parameters",
                                          sections = ["castep_section_atom_positions"],
                                          subMatchers = [

                                            SM(r"\s*a \=\s*(?P<castep_cell_length_a>[\d\.]+)\s*alpha \=\s*(?P<castep_cell_angle_alpha>[\d\.]+)"),
                                            SM(r"\s*b \=\s*(?P<castep_cell_length_b>[\d\.]+)\s*beta  \=\s*(?P<castep_cell_angle_beta>[\d\.]+)"),
                                            SM(r"\s*c \=\s*(?P<castep_cell_length_c>[\d\.]+)\s*gamma \=\s*(?P<castep_cell_angle_gamma>[\d\.]+)"),
                                            SM(r"\s*x\s*(?P<castep_store_atom_labels>[A-Za-z0-9]+\s+[\d\.]+)\s*[0-9]\s*(?P<castep_store_atom_positions>[\d\.]+\s+[\d\.]+\s+[\d\.]+)",
                                               endReStr = "\n",
                                               repeats = True)

                                          ]), # CLOSING castep_section_atom_positions


                                       # SCF Single Point Evaluation (energies and forces)
                                       SM(name = 'SinglePointEvaluation',
                                          startReStr = r"SCF\sloop\s*Energy\s*Energy\sgain\s*Timer\s*<\-\-\sSCF\s*",
                                          #repeats = True,
                                          forwardMatch = True,
                                          sections = ['section_single_configuration_calculation'],
                                          subMatchers = [

                                              SM(name = 'ScfIterations',
                                                 startReStr = r"SCF\sloop\s*Energy\s*Energy\sgain\s*Timer\s*<\-\-\sSCF\s*",
                                                 sections = ['section_scf_iteration'],
                                                 subMatchers = [

                                                    SM(r"\s*[0-9]+\s*(?P<energy_total_scf_iteration>[-+0-9.eEdD]*)\s*[-+0-9.eEdD]*\s*[0-9.]*\s*\<\-\-\sSCF\s*",
                                                       repeats = True),

                                                 ]), # CLOSING section_scf_iteration


                                          SM(r"Final energy = *(?P<energy_total>[-+0-9.eEdD]*)"), # macthing final coverged total energy

                                          #  Band Structure Calculation: here we match and parse k points and eigenvalues
                                          SM(startReStr = "\s*\+\s*B A N D   S T R U C T U R E   C A L C U L A T I O N\s*",
                                                 forwardMatch = True,
                                                 sections = ["castep_section_k_band_group"],
                                                 subMatchers = [

                                                     # First spin channel
                                                     SM(startReStr = "\s*\+\s*Spin\=1\skpt\=\s*",
                                                        sections = ["castep_section_k_band"],
                                                        forwardMatch = True,
                                                        subMatchers = [

                                                         SM(startReStr = "\s*\+\s*Spin\=1\skpt\=\s*",
                                                            sections = ["castep_section_k_points"],
                                                            forwardMatch = True,
                                                            repeats = True,
                                                            subMatchers = [
                                                                # Matching k points
                                                                SM(r"\s*\+\s*Spin\=1\s*kpt\=\s*[0-9]+\s*\((?P<castep_store_k_points>\s+[-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)\)\s*",
                                                                   repeats = True),

                                                                    SM(name = 'Eigen_1',
                                                                       startReStr = r"\s*\+\s*\+\s*",
                                                                       sections = ['castep_section_eigenvalues'],
                                                                       repeats = True,
                                                                       subMatchers = [
                                                                          # Matching eigenvalues
                                                                          SM(r"\s*\+\s*[0-9]+\s*(?P<castep_store_eigenvalues>\s+[-\d\.]+)",
                                                                             repeats = True)

                                                                       ]), # CLOSING castep_section_eigenvalues


                                                            ]), # CLOSING castep_section_k_points


                                                        ]), # CLOSING 1st section_eigenvalues

                                                     # Second spin channel
                                                     SM(startReStr = "\s*\+\s*Spin\=2\skpt\=\s*",
                                                        sections = ["castep_section_k_band"],
                                                        forwardMatch = True,
                                                        subMatchers = [

                                                         SM(startReStr = "\s*\+\s*Spin\=2\skpt\=\s*",
                                                            sections = ["castep_section_k_points_1"],
                                                            forwardMatch = True,
                                                            repeats = True,
                                                            subMatchers = [
                                                                # Matching k points
                                                                SM(r"\s*\+\s*Spin\=2\s*kpt\=\s*[0-9]+\s*\((?P<castep_store_k_points_1>\s+[-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)\)\s*",
                                                                   repeats = True),

                                                                    SM(name = 'Eigen_2',
                                                                       startReStr = r"\s*\+\s*\+\s*",
                                                                       sections = ['castep_section_eigenvalues_1'],
                                                                       repeats = True,
                                                                       subMatchers = [
                                                                          # Matching eigenvalues
                                                                          SM(r"\s*\+\s*[0-9]+\s*(?P<castep_store_eigenvalues_1>\s+[-\d\.]+)",
                                                                             repeats = True)

                                                                       ]), # CLOSING castep_section_eigenvalues_1


                                                            ]), # CLOSING castep_section_k_points_1


                                                        ]), # CLOSING 2nd section_eigenvalues


                                                ]), # CLOSING section_eigenvalues_group


                                          SM(startReStr = r"\s\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\* Symmetrised Forces \*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\*\s*",
                                             subMatchers = [
                                                SM(r"\s*\*\s*[A-Za-z]+\s*[0-9]\s*(?P<castep_store_atom_forces>[-\d\.]+\s+[-\d\.]+\s+[-\d\.]+)", repeats = True)
                                             ])


                                          ]), # CLOSING SM SinglePointEvaluation


                                ]), # CLOSING section_system


    ]), # CLOSING SM newRun
])






# loading metadata from nomad-meta-info/meta_info/nomad_meta_info/castep.nomadmetainfo.json
metaInfoPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../../../../nomad-meta-info/meta_info/nomad_meta_info/castep_old.nomadmetainfo.json"))
metaInfoEnv, warnings = loadJsonFile(filePath = metaInfoPath, dependencyLoader = None, extraArgsHandling = InfoKindEl.ADD_EXTRA_ARGS, uri = None)
# dictionary for functions to call on close of a section metaname
onClose = {}
# parser info
parserInfo = {'name':'castep-parser', 'version': '1.0'}
# adjust caching of metadata
cachingLevelForMetaName = {'energy_total': CachingLevel.Cache,
                           'energy_total_scf_iteration_list': CachingLevel.Forward,
                           'atom_forces': CachingLevel.Forward,
                           'atom_labels': CachingLevel.Forward,
                           'number_of_atoms': CachingLevel.Forward,
                           'castep_atom_positions': CachingLevel.Forward,
                           'atom_positions': CachingLevel.Forward,
                           'basis_set_planewave_cutoff': CachingLevel.Forward,
                           'castep_basis_set_planewave_cutoff': CachingLevel.Cache,
                           'basis_set_cell_dependent_kind': CachingLevel.Forward,
                           'basis_set_cell_dependent_name': CachingLevel.Forward,

                           'castep_band_kpoints': CachingLevel.Forward,
                           'castep_band_energies': CachingLevel.Forward,
                           'castep_number_of_band_kpoints': CachingLevel.Forward,
                           'castep_number_of_band_energies': CachingLevel.Forward,
                           'castep_store_k_points': CachingLevel.Cache,
                           'castep_store_eigenvalues': CachingLevel.Cache,
                           'castep_store_k_points_1': CachingLevel.Cache,
                           'castep_store_eigenvalues_1': CachingLevel.Cache,
                           'castep_store_atom_forces': CachingLevel.Cache,
                           'castep_store_atom_labels': CachingLevel.Cache,
                           }

if __name__ == "__main__":
    mainFunction(mainFileDescription, metaInfoEnv, parserInfo, superContext = CastepParserContext(), cachingLevelForMetaName = cachingLevelForMetaName, onClose = onClose,
                 defaultSectionCachingLevel = True)
